/*----------------------------------------------------------------*
|tfmcc.h                                                          |
|@version: 2.0                                                    |
|Last Modification: Dec 14th, 2003 , by Fei, based on 0516032252  |
|                                                                 |
| Define TFMCC threads and functions                              |
*----------------------------------------------------------------*/
/*----------------------------------------------------------------*
|    Copyright 2001-2003 Networking and Simulation Laboratory     |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for academic purposes is hereby  |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
 *---------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#ifdef UNIX
#include <pthread.h>
#else
#include <winsock.h>
#include <windows.h>
#include <process.h>
#endif
#include <math.h>
#include <sys/timeb.h>
#include <fstream.h>
/* constant, data strucutre and function prototype for TFMCC */
#include "tfmcc_proto.h"

/*****************************************************************
               global variables
******************************************************************/
/* table of multicast addresses 
 * synchronized by tfmcc_mca_mutex      
 */
unsigned long int tfmcc_mca[20];
#ifdef UNIX
pthread_mutex_t tfmcc_mca_mutex;
#else
HANDLE tfmcc_mca_mutex;
#endif

/* for each mutlicast group address, there is a record */
struct group_info_t * group_info = new struct group_info_t[20];


/* The following variables are for printing out the packet
 * loss ratio at the receiver side
 */
int r_time_axis=-1;
FILE * LossRecord;
unsigned short int r_start;
unsigned short int r_stop;
unsigned short int r_delay;
int number_packet_received=0;
int number_packet_lost=0;
double packet_loss_ratio;

/********************************************************************************
                          functions
*********************************************************************************/
/*******************************************************************************
* The function to initialize all the TFMCC data structures                     *
* TFMCC threds are not created here. They are created by search_tfmcc_mca()    *
*   when SRMP subscribe or register a new multicast group                      *
********************************************************************************/
void initialize_tfmcc()
{  int i;
   int j;
   int k;

   /* get local ip address, and set the sender_id and receiver_id */
#ifdef UNIX
   char host_name[100];
   struct hostent* host_ent;
   struct in_addr local_addr;

   gethostname(host_name, sizeof(host_name));
   host_ent=gethostbyname(host_name);
   memcpy(&local_addr.s_addr, host_ent->h_addr_list[0], host_ent->h_length);

   strcpy(ip_address, inet_ntoa(local_addr));
#else
   char szHostname[100];
   WSADATA wsaData;
   HOSTENT *pHostEnt;
   int nAdapter = 0;
   struct in_addr sAddr;
   if (WSAStartup(0x0101, &wsaData))
   {  printf("WSAStartup failed %s\n", WSAGetLastError());
      exit(-1);
   }
   gethostname( szHostname, sizeof( szHostname ));
   pHostEnt = gethostbyname( szHostname );
   memcpy ( &sAddr.s_addr, pHostEnt->h_addr_list[nAdapter],
                  pHostEnt->h_length);

   /* print the machines IP Address. */
   strcpy(ip_address, inet_ntoa(sAddr));
#endif
   printf("TFMCC V2.0\n");

   /* mutex for tfmcc_mca[] */
#ifdef UNIX
   pthread_mutex_init(&tfmcc_mca_mutex,NULL);
#else
   tfmcc_mca_mutex=CreateMutex(NULL, FALSE, "tfmcc_mca lock");
#endif

   /* initialize data structures */
   for(k=0; k<20; k=k+1)
   {  memset(&group_info[k], 0, sizeof(struct group_info_t));

      for(i=0; i<20; i=i+1)
      {  group_info[k].srmp_send_his[i].len=0;
         group_info[k].srmp_send_his[i].mode=0;
         group_info[k].srmp_send_his[i].sending_time=65535;
         group_info[k].mca=0;
      }
#ifdef UNIX
      pthread_mutex_init(&(group_info[k].mutex),NULL);
      pthread_cond_init(&(group_info[k].event), NULL);
      pthread_mutex_init(&(group_info[k].other_senders_mutex), NULL);
      pthread_mutex_init(&(group_info[k].time_event_mutex), NULL);
      pthread_cond_init(&(group_info[k].time_event), NULL);
      pthread_cond_init(&(group_info[k].fbnr_event), NULL);
      pthread_mutex_lock(&(group_info[k].mutex));
#else
      group_info[k].mutex=CreateMutex(NULL, FALSE, "group_info lock");
      group_info[k].event=CreateEvent(NULL, FALSE, FALSE, "non_clr feedback received");
      group_info[k].other_senders_mutex=CreateMutex(NULL, FALSE, "receiver_info lock");
      if(group_info[k].other_senders_mutex==0)
      {  printf("error create mutex\a\n");
         exit(-1);
      }
      group_info[k].time_event_mutex=CreateMutex(NULL, FALSE, "time event lock");
      group_info[k].time_event=CreateEvent(NULL, TRUE, FALSE, "timer event");
      group_info[k].fbnr_event=CreateEvent(NULL,FALSE,FALSE, "rmax round event");  
      if( group_info[k].fbnr_event == NULL )
         printf("fbnr_event creation failed\n");

      WaitForSingleObject(group_info[k].mutex, INFINITE);
#endif

      group_info[k].feedback_listhead=NULL;
      strcpy(group_info[k].sender_id, ip_address);

      group_info[k].x_send=-1; /* -1 means 1 packet per feedback round */

      group_info[k].r_max=INITIAL_RMAX;
      group_info[k].x_supp=(double)1e30;

      group_info[k].no_large_rmax=1;

      group_info[k].clr_new_flag=0;
      group_info[k].in_slowstart_flag=1;
      group_info[k].fb_nr=0;
      group_info[k].feedback_received_flag=0;
      /* flag that indicate if a larger RTT has been received during a feedback round */
      group_info[k].non_feedback=0; 

      group_info[k].target_rate=-1;
      group_info[k].start_point=65535;
      group_info[k].increase_rate=-1;

      group_info[k].feedback_rate=-1;
      group_info[k].slowstart_rate=-1;
      group_info[k].cancel_flag=0;
      group_info[k].current_receiver_leave=0;
      group_info[k].mean_packet_size=-1;

#ifdef UNIX
      pthread_mutex_unlock(&(group_info[k].mutex));
      pthread_mutex_lock(&(group_info[k].other_senders_mutex));
#else
      ReleaseMutex(group_info[k].mutex);
      WaitForSingleObject(group_info[k].other_senders_mutex, INFINITE);
#endif
      for(i=0; i<20; i=i+1)
      {  memset(&group_info[k].other_senders[i], 0, sizeof(struct other_sender_t));
         group_info[k].other_senders[i].s_old=-1;
         group_info[k].other_senders[i].have_loss_flag=0;
         group_info[k].other_senders[i].minimum_x_supp=1e20;
         strcpy(group_info[k].other_senders[i].receiver_id, ip_address);
         group_info[k].other_senders[i].sender_list.fb_nr=-1;
         group_info[k].other_senders[i].mean_packet_size=-1;
         group_info[k].other_senders[i].number_of_loss=0;

         for(j=0; j<20; j=j+1)
         {  group_info[k].other_senders[i].packet_history[j].sn=-1;
            group_info[k].other_senders[i].packet_history[j].received_time=65535;
         }
      }

      group_info[k].time_event_head=NULL;
      group_info[k].time_event_tail=NULL;
#ifdef UNIX
      pthread_mutex_unlock(&(group_info[k].other_senders_mutex));
#else
      ReleaseMutex(group_info[k].other_senders_mutex);
#endif
   }
   LossRecord=fopen("loss.txt","w");
   fprintf(LossRecord, "TitleText: Packet Loss Ratio at TFMCC receiver\n");
   fprintf(LossRecord, "\" loss ratio at %s (in Percent)\n", ip_address);
}  /* end of function initialize_tfmcc() */


/**************************************************************************************
* The function to generate a timestamp                                                *
* The timestamp value is between 0 and 65535                                          *
***************************************************************************************/
unsigned short int get_timestamp()
{
#ifdef UNIX
   struct timeb timebuffer;
#else
   struct _timeb timebuffer;
#endif
   unsigned long int ticks;
   unsigned short int timestamp;

#ifdef UNIX
   ftime( &timebuffer );
#else
   _ftime( &timebuffer );
#endif
   ticks=timebuffer.millitm+timebuffer.time*1000;
   timestamp=(unsigned short)ticks;

   return timestamp;
}

/**************************************************************************************
 * This function should be called by SRMP, and return a bundle_header_para.           *
 * After execution, the bundler_header parameter will contain all the information to  *
 * compose a bundle header, except dsn_count and len.                                 *
 *     0. update the mean_packet_size for the sender of the specified mca             *
 *     1. fill in the bundle_header                                                   *
 *     2. update the feedback_list_head, to remove all obsolete records               *
 *     3. update the sending rate                                                     *
 *     4. choose a receiver according to TFMCC draft section 3.5                      *
 **************************************************************************************/
int upon_send_bundle( struct bundle_header_t * bundler_header,
                     double * sending_rate, double packet_size, unsigned long int mca)
{  struct feedback_list_t * temp;  /* temp variable used to search the link list */
   struct feedback_list_t * temp2=NULL;
   double rate;
   unsigned short int current_time;
   int group_ind;
   group_ind=search_tfmcc_mca(mca,0);

   /***********************************/
   /*    update the mean packet size  */
   /***********************************/
   memset(bundler_header, 0, sizeof( struct bundle_header_t));
#ifdef UNIX
   pthread_mutex_lock(&(group_info[group_ind].mutex));
#else
   WaitForSingleObject(group_info[group_ind].mutex, INFINITE);
#endif
   if(group_info[group_ind].mean_packet_size<0)
      group_info[group_ind].mean_packet_size=packet_size;
   else
   {  group_info[group_ind].mean_packet_size=
            group_info[group_ind].mean_packet_size+(double)packet_size;
      group_info[group_ind].mean_packet_size=group_info[group_ind].mean_packet_size/2;
   }

   /***********************************/
   /*    fill in the bundle header    */
   /***********************************/
   strcpy(bundler_header->sender_id, group_info[group_ind].sender_id);
   bundler_header->x_supp=group_info[group_ind].x_supp;
   bundler_header->r_max=group_info[group_ind].r_max;
   bundler_header->fb_nr=group_info[group_ind].fb_nr;
   bundler_header->sn=group_info[group_ind].sn;

   group_info[group_ind].sn=group_info[group_ind].sn+1;
   if(group_info[group_ind].sn>65535)
      group_info[group_ind].sn=0;

   bundler_header->receiver_timestamp=0;
   bundler_header->is_clr_flag=0;

   /* note, in the receiver, a variable to indicate measuring, 
    * set measuring =0 will remove all old feedback
    */

   /***********************************/
   /*    fill in the bundle header    */
   /***********************************/
   /* check if there is obsolete entry in feedback_listhead. If yes, remove them
    * that is, fb_nr < current fb_nr.
    * If the receive_id of the feedback_list equals to that of CLR,
    * which means the receiver has newly became the CLR.
    * In this case, the feedback_list should be removed.
    */
   temp=group_info[group_ind].feedback_listhead;
   temp2=NULL;
   while( temp!=NULL)
   {  if( temp->fb_nr!=group_info[group_ind].fb_nr ||
                   strcmp(temp->receiver_id, group_info[group_ind].clr.receiver_id)==0)
      {  /* remove the entry from link list*/
         if(temp->prev ==NULL)
            group_info[group_ind].feedback_listhead=temp->next;
         else
            temp->prev->next=temp->next;

         if(temp->next!=NULL)
            temp->next->prev=temp->prev;
         else
            group_info[group_ind].feedback_listtail=temp->prev;

         temp2=temp;
      }

      temp=temp->next;
      if(temp2!=NULL)
         free(temp2);
      temp2=NULL;
   }

   /********************************/
   /*   update the sending rate    */
   /********************************/
   if(group_info[group_ind].x_send < group_info[group_ind].target_rate)
   {  if(group_info[group_ind].x_send<0)
         group_info[group_ind].x_send=0;
      current_time=get_timestamp();
      if(group_info[group_ind].in_slowstart_flag==1)
      {  rate=group_info[group_ind].x_send+
               (current_time-group_info[group_ind].start_point)
               *group_info[group_ind].increase_rate;
         /* printf("       increase rate=%f\n", group_info[group_ind].increase_rate); */
      }
      else
      {  if(group_info[group_ind].mean_packet_size<0)
            rate=group_info[group_ind].x_send+
                 (current_time-group_info[group_ind].start_point)
                 *8*DEFAULT_MEAN_PKT_SIZE/
                 group_info[group_ind].r_max/group_info[group_ind].r_max*1000;
         else
            rate=group_info[group_ind].x_send+
                 (current_time-group_info[group_ind].start_point)
                 *8*group_info[group_ind].mean_packet_size/
                 group_info[group_ind].r_max/group_info[group_ind].r_max*1000;
      }

      group_info[group_ind].start_point=current_time;

      if(rate>=group_info[group_ind].target_rate)
      {  rate=group_info[group_ind].target_rate;
         group_info[group_ind].increase_rate=0;
         group_info[group_ind].target_rate=-1;
         group_info[group_ind].start_point=65535;
      }

      group_info[group_ind].x_send=rate;
   }

   *sending_rate=group_info[group_ind].x_send;


   /*****************************************/
   /*  choose a receiver to echo timestamp  */
   /*****************************************/
   /* as specified by chapter 3.5 of tfmcc draft */

   /* First if a new CLR, or the CLR have not measured its RTT */
   if (strlen(group_info[group_ind].clr.receiver_id)>0)
   {  if ((group_info[group_ind].clr_new_flag==1) || 
            (group_info[group_ind].clr.have_rtt_flag==0))
       /* note: have_rtt_flag for CLR may still be 0, when previous echoed measure 
        * have not reached CLR. So, may need a flag here.
        */
      {  /* set bundler header field */
         strcpy(bundler_header->receiver_id, group_info[group_ind].clr.receiver_id);
         bundler_header->receiver_timestamp = 
             group_info[group_ind].clr.receiver_timestamp+
             (get_timestamp() - group_info[group_ind].clr.received_time);
         bundler_header->is_clr_flag=1;

         /* unset group_info */
         group_info[group_ind].clr_new_flag=0;
         group_info[group_ind].clr.have_rtt_flag=1; 
         bundler_header->sender_timestamp=get_timestamp();
         /* show_bundle_header(bundler_header); */
         /* printf("      sending rate= %f\n", *sending_rate); */
#ifdef UNIX
         pthread_mutex_unlock(&(group_info[group_ind].mutex));
#else
         ReleaseMutex(group_info[group_ind].mutex);
#endif
         return 0;
      }
   }

   /* Second, else, the oldest feedback that have not measured its RTT */
   temp=group_info[group_ind].feedback_listhead;
   while( temp!=NULL)
   {  if( temp->have_rtt_flag==0)
      {  strcpy(bundler_header->receiver_id, temp->receiver_id);
         bundler_header->receiver_timestamp = temp->receiver_timestamp+
             (get_timestamp() - temp->received_time);

         /* remove the entry from link list */
         if(temp->prev ==NULL)
            group_info[group_ind].feedback_listhead = temp->next;
         else
            temp->prev->next=temp->next;

         if(temp->next!=NULL)
            temp->next->prev=temp->prev;
         else
            group_info[group_ind].feedback_listtail=temp->prev;

         free(temp);
         /* note: this value should be set in SRMP for accuracy */
         bundler_header->sender_timestamp=get_timestamp();
         /* show_bundle_header(bundler_header); */
         /* printf("      sending rate= %f\n", *sending_rate); */
#ifdef UNIX
         pthread_mutex_unlock(&(group_info[group_ind].mutex));
#else
         ReleaseMutex(group_info[group_ind].mutex);
#endif         
         return 0;
      }
      temp=temp->next;
   }

   /* Third, else, the oldest NONE CLR feedback; */
   temp=group_info[group_ind].feedback_listhead;
   while( temp!=NULL)
   {  if(strcmp(temp->receiver_id, group_info[group_ind].clr.receiver_id)!=0)
      {  strcpy(bundler_header->receiver_id, temp->receiver_id);
         bundler_header->receiver_timestamp = temp->receiver_timestamp+
                (get_timestamp() - temp->received_time);

         /* remove the entry from link list */
         if(temp->prev ==NULL)
            group_info[group_ind].feedback_listhead=temp->next;
         else
            temp->prev->next=temp->next;

         if(temp->next!=NULL)
            temp->next->prev=temp->prev;
         else
            group_info[group_ind].feedback_listtail=temp->prev;

         free(temp);
         /* note: this value should be set in SRMP for accuracy */
         bundler_header->sender_timestamp=get_timestamp();
         /* show_bundle_header(bundler_header); */
         /* printf("      sending rate= %f\n", *sending_rate); */
#ifdef UNIX
         pthread_mutex_unlock(&(group_info[group_ind].mutex));
#else
         ReleaseMutex(group_info[group_ind].mutex);
#endif
         return 0;
      }
      temp=temp->next;
   }


    /* Fourth, if none of the above, the CLR */
   if (strlen(group_info[group_ind].clr.receiver_id)>0)
   {  strcpy(bundler_header->receiver_id, group_info[group_ind].clr.receiver_id);
      bundler_header->receiver_timestamp=
            group_info[group_ind].clr.receiver_timestamp+
            (get_timestamp() - group_info[group_ind].clr.received_time);
      bundler_header->is_clr_flag=1;
      /*  note: this value should be set in SRMP for accuracy */
      bundler_header->sender_timestamp=get_timestamp();
      /* show_bundle_header(bundler_header); */
      /* printf("      sending rate= %f\n", *sending_rate); */
#ifdef UNIX
      pthread_mutex_unlock(&(group_info[group_ind].mutex));
#else
      ReleaseMutex(group_info[group_ind].mutex);
#endif
      return 0;
   }

   /* all the above failed. That means: No CLR, no feedback during last feedback round.
    * this may happy during the initialize stage. note: However, the draft does not say
    * anything about this
    * note: this value should be set in SRMP for accuracy
    */
   bundler_header->sender_timestamp=get_timestamp();
   /* show_bundle_header(bundler_header); */
   /* printf("      sending rate= %f\n", *sending_rate); */
#ifdef UNIX
   pthread_mutex_unlock(&(group_info[group_ind].mutex));
#else
   ReleaseMutex(group_info[group_ind].mutex);
#endif
   return 0;

}


/**************************************************************************************/
/* This function should be called by SRMP, when each time a feedback is received      */
/*   0. update current sending_rate, but not the target sending_rate                  */
/*   If the feedback is for local sender, then                                        */
/*       1. calculate RTT                                                             */
/*       If it is not from CLR, then                                                  */
/*            2. triger a new feedback round if possible                              */
/*            3. insert the receiver information into feedback_list head              */
/*            4. update the supression rate                                           */
/*         5. check if slow start stage should be terminated                          */
/*         6. update r_max                                                            */
/*         7. update CLR according to TFMCC draft section 3.3, and set sending rate   */
/*     Else, the feedback is not for local sender                                     */
/**************************************************************************************/
int upon_receive_feedback(struct feedback_header_t feedback_header1,
                              unsigned long int mca)
{  /* note to updata current sending rate each time received a feedback message */
   int index;
   unsigned short int rtt;
   struct feedback_list_t * temp;
   /* searh in group_info[] to find the correct index */
   struct feedback_header_t feedback_header;
   unsigned short int current_time;
   double rate;
   int i;
   int group_ind;
   /* show_feedback_header(feedback_header1); */
   group_ind=search_tfmcc_mca(mca, 0);
#ifdef UNIX
   pthread_mutex_lock(&(group_info[group_ind].mutex));
#else
   WaitForSingleObject( group_info[group_ind].mutex, INFINITE);
#endif
   /*********************************/
   /*  update current sending rate  */
   /*********************************/
   /* update current sending rate, only to update the current sending rate,
    * but not to set the target rate etc.
    */
   if(group_info[group_ind].x_send < group_info[group_ind].target_rate)
   {  if(group_info[group_ind].x_send<0)
         group_info[group_ind].x_send=0;

      current_time=get_timestamp();
      if(group_info[group_ind].in_slowstart_flag==1 )
      {  rate=group_info[group_ind].x_send+
               (current_time-group_info[group_ind].start_point)
               *group_info[group_ind].increase_rate;
      }
      else
      {  rate=group_info[group_ind].x_send+
               (current_time-group_info[group_ind].start_point)
               *8*group_info[group_ind].mean_packet_size
               /group_info[group_ind].r_max/group_info[group_ind].r_max*1000;
      }
      /*printf("rate=%f\n",
      8*group_info[group_ind].mean_packet_size
      /group_info[group_ind].r_max/group_info[group_ind].r_max*1000);*/
      group_info[group_ind].start_point=current_time;

      if(rate>group_info[group_ind].target_rate)
      {  rate=group_info[group_ind].target_rate;
         group_info[group_ind].increase_rate=0;
         group_info[group_ind].target_rate=-1;
         group_info[group_ind].start_point=65535;
      }

      group_info[group_ind].x_send=rate;
   }
   /* printf("      sending rate=%f\n", group_info[group_ind].x_send); */

   memcpy(&feedback_header, &feedback_header1, sizeof(struct feedback_header_t));
   /* printf("        desired_rate=%f\n",feedback_header.x_r); */

   /***********************************************/
   /*  check if the feedback is for local sender  */
   /***********************************************/
   if(strcmp(feedback_header.sender_id, group_info[group_ind].sender_id)==0)
   {  /* set the feedback_received_flag, to indicate that a non_clr feedback has been
       * received during a feedback round, this flag will be reset by the fb_nr thread
       */
      /*printf("feedback received from: %s\n", feedback_header.receiver_id); */

      group_info[group_ind].non_feedback=0; /* any kind of feedback received */

   /* check to see if the feedback already obsolete. If yes, do nothing but return.
      if(feedback_header.fb_nr != group_info[group_ind].fb_nr)
      {  ReleaseMutex(group_info[group_ind].mutex);
         printf("       obsoleted Feedback received and discarded\a\n");
         return 0;
      }
   */
      /* some times the received time is less than timestamp echoed. This may be 
         caused by either timestamp wrapped, or the system clock incurracy */

      /*******************/
      /*  calculate RTT  */
      /*******************/
      if(
        (feedback_header.received_time < feedback_header.sender_timestamp)
        &&
        ((feedback_header.sender_timestamp-feedback_header.received_time) <32768)
        )
      {  rtt=0+SIMU_DELAY;
      }
      else
      {  rtt=(feedback_header.received_time - feedback_header.sender_timestamp)
               + SIMU_DELAY;
      }
      /* printf("   rtt=%d\n", rtt); */


      if(strcmp(group_info[group_ind].clr.receiver_id,feedback_header.receiver_id)!=0 )
      {  /* non_clr feedback received */
         group_info[group_ind].feedback_received_flag=1;

        /*********************************************/
        /*  triger a new feedback round if possible  */
        /*********************************************/
#ifdef UNIX
        pthread_cond_signal(&(group_info[group_ind].event));
#else
        SetEvent(group_info[group_ind].event);
#endif
        /**********************************************/
        /* insert the feedback into feedback_listhead */
        /**********************************************/
        temp = (struct feedback_list_t *)malloc(sizeof(struct feedback_list_t));
        memset(temp, 0, sizeof(struct feedback_list_t));
        if(temp==NULL)
        {  printf("malloc error!\n");
           exit(-1);
        }
         memset(temp, 0, sizeof(struct feedback_list_t));
        strcpy(temp->receiver_id,feedback_header.receiver_id);
        /*note: the following line is for test harness, 
                because it can't provide accurate received_time */
        /* temp->received_time=feedback_header.received_time; */
        temp->received_time=get_timestamp();
        temp->rtt = rtt;
        temp->have_rtt_flag=feedback_header.have_rtt_flag;
        temp->have_loss_flag=feedback_header.have_loss_flag;
        temp->receiver_leave_flag=feedback_header.receiver_leave_flag;
        temp->fb_nr=feedback_header.fb_nr;
        temp->receiver_timestamp=feedback_header.receiver_timestamp;

        if(group_info[group_ind].feedback_listtail==NULL)
        {  group_info[group_ind].feedback_listhead=temp;
           temp->prev=NULL;
        }
        else
        {  temp->prev=group_info[group_ind].feedback_listtail;
           group_info[group_ind].feedback_listtail->next=temp;
        }
        group_info[group_ind].feedback_listtail=temp;
        temp->next=NULL;
        /********************************/
        /*  update the supression rate  */
        /********************************/
        if(feedback_header.x_r < group_info[group_ind].x_supp)
          group_info[group_ind].x_supp=0.9*feedback_header.x_r;

      }
      /* note: rember to delte entry in feedback_list that equeals to current CLR,
         when sent out bundle */


      /*****************************************************/
      /*  check to see if slow start should be terminated  */
      /*****************************************************/
      if(group_info[group_ind].in_slowstart_flag==1 && 
           feedback_header.have_loss_flag==1)
      {  group_info[group_ind].in_slowstart_flag=0;
         printf("      **** slowstart terminated ****\n");
         group_info[group_ind].slowstart_rate=-1;
         group_info[group_ind].target_rate=-1;
         group_info[group_ind].increase_rate=0;
         /* note to check if should set the data rate to the desired rate 
           reported in the feedback */
      }

      /* calculate RTT, and check to see if R_max should be updated */
      /*rtt=(feedback_header.received_time - feedback_header.receiver_timestamp)/1000;*/

      /***********************/
      /*    update r_max     */
      /***********************/
      if(rtt> group_info[group_ind].r_max)
      {  group_info[group_ind].r_max=rtt;
         group_info[group_ind].no_large_rmax=0;  
        /* note, this flag should be reset somewhere */
        /* note, in fbnr thread,
           rember to update R_max based on no_large_rmax */
      }
      /*else if(no_large_rmax ==0)
            no_large_rmax=1;*/
      group_info[group_ind].r_max=
         (unsigned short int) max_value(group_info[group_ind].r_max, 
              group_info[group_ind].mean_packet_size/group_info[group_ind].x_send);
      /* note ,string pointer will never be NULL, 
       * ie, group_info[group_ind].clr!=NULL 
       */

      /*
       * reset non_feedback
       * note:int fb_nr thread, should count this value to 10 to halve the send rate
       * if(feedback_header.receiver_id!=group_info[group_ind].clr.receiver_id)
       *       group_info[group_ind].non_feedback=0;
       */



      /* Calculate the new sending rate*/
      /* note the draft does not specify what is the "receiver with the smallest report
       * rate". Maybe, the receiver whose maximum report rate is the smallest among
       * all receivers or else, during a feedback round, the receiver with the
       * smallest report rate. Other wise the first receiver feedback  with 2 
       * packets/feedback round will not increase.
       */
      if(group_info[group_ind].in_slowstart_flag ==1)
      {  /* in slow start, first check if the new feedback will leads to the change of
          * target clr. Slowstart_rate is the target sending rate, during slow start, 
          * and will take effect right after the old feedback round terminated and a
          * new one started.
          */
        if(group_info[group_ind].slowstart_rate<0 ||
          feedback_header.x_r<group_info[group_ind].slowstart_rate)
        {  group_info[group_ind].slowstart_rate=feedback_header.x_r;

           /* note, what if the feedback is from current CLR?
            * the draft didn't specify that.
            */
           strcpy(group_info[group_ind].target_clr.receiver_id,
                  feedback_header.receiver_id);
           group_info[group_ind].target_clr.receiver_leave_flag=
                  feedback_header.receiver_leave_flag;
           /* note: the following line is for test harness, because it can't provice
               accurate received_time */
           /*group_info[group_ind].target_clr.received_time=
                                                   feedback_header.received_time; */
           group_info[group_ind].target_clr.received_time=get_timestamp();
           group_info[group_ind].target_clr.rtt = rtt;
           group_info[group_ind].target_clr.have_rtt_flag=feedback_header.have_rtt_flag;
           group_info[group_ind].target_clr.have_loss_flag=
                                                        feedback_header.have_loss_flag;
           group_info[group_ind].target_clr.receiver_leave_flag=
                                                   feedback_header.receiver_leave_flag;
           group_info[group_ind].target_clr.receiver_timestamp=
                                                    feedback_header.receiver_timestamp;


           /* return 0; */
        }
        else if( strcmp(group_info[group_ind].target_clr.receiver_id, 
                           feedback_header.receiver_id)==0)
        {  if(feedback_header.x_r>group_info[group_ind].slowstart_rate)
           {  group_info[group_ind].slowstart_rate=feedback_header.x_r;
           }
        }



        /* in slow start, second, if the feedback is for CLR, update CLR information */
        if(strcmp(group_info[group_ind].clr.receiver_id, feedback_header.receiver_id)
             ==0)
        {  /*group_info[group_ind].clr.receiver_id=feedback_header.receiver_id;*/
           group_info[group_ind].clr.receiver_timestamp=
                                          feedback_header.receiver_timestamp;
           /* note:the following line is for test harness,
                  because it can't provide accurate received tim */
           /* group_info[group_ind].clr.received_time=feedback_header.received_time; */
           group_info[group_ind].clr.received_time=get_timestamp();
           group_info[group_ind].clr.receiver_leave_flag=
                                          feedback_header.receiver_leave_flag;
           /*group_info[group_ind].clr_new_flag=1; */
           group_info[group_ind].clr_non_feedback=0;
           group_info[group_ind].clr_leave_flag=feedback_header.receiver_leave_flag;
           group_info[group_ind].clr.have_loss_flag=feedback_header.have_loss_flag;
           group_info[group_ind].clr.have_rtt_flag=feedback_header.have_rtt_flag;

        }
#ifdef UNIX
        pthread_mutex_unlock(&(group_info[group_ind].mutex));
#else
        ReleaseMutex(group_info[group_ind].mutex);
#endif
        /* update the mimum x_r from all the receivers
         * this is for feedback suppression
         * note, should wait for some mutex here.
         * search for the index of the sender_id,
         * if does not exist, create one;
         */
/*      WaitForSingleObject(group_info[group_ind].other_senders_mutex, INFINITE);
        for(i=0; i<20; i=i+1)
        {  if(strcmp(group_info[group_ind].other_senders[i].sender_list.sender_id,
                              feedback_header.sender_id)==0)
           {  index=i;
              break;
           }
        }

        if(i<20 && strcmp(feedback_header.receiver_id, 
                           group_info[group_ind].other_senders[index].clr_id)!=0)
        {  if(feedback_header.x_r<
                        group_info[group_ind].other_senders[index].minimum_x_supp)
           group_info[group_ind].other_senders[index].minimum_x_supp = 
                        feedback_header.x_r;
        }
        ReleaseMutex(group_info[group_ind].other_senders_mutex);

*/

        return 0;
      }

      /* if it is not in slow start stage */
      if(feedback_header.have_rtt_flag==0)
        feedback_header.x_r=feedback_header.x_r * group_info[group_ind].r_max / rtt;

      /* there is no current CLR, the receiver that send the feedback became CLR.
       * Draft 3.3-1
       */
      if(strlen(group_info[group_ind].clr.receiver_id)==0)
      {  strcpy(group_info[group_ind].clr.receiver_id, feedback_header.receiver_id);
         group_info[group_ind].clr.receiver_timestamp=
                           feedback_header.receiver_timestamp;
         /* note:the following line is for test harness,
          * because it can't provide accurate received time
          */
         /*group_info[group_ind].clr.received_time=feedback_header.received_time;*/
         group_info[group_ind].clr.received_time=get_timestamp();
         group_info[group_ind].clr.receiver_leave_flag= 
                                             feedback_header.receiver_leave_flag;
         group_info[group_ind].clr_new_flag=1;
         group_info[group_ind].clr_non_feedback=0;
         group_info[group_ind].clr_leave_flag=feedback_header.receiver_leave_flag;
         /* note to check if non_feedback should also be reset,
          * to indicate that NON-CLR feedback received
          * note to check if ... flag should also be reset, to indicate that 
          * feedback received
          */


        if(feedback_header.x_r<(group_info[group_ind].x_send+
           8*group_info[group_ind].mean_packet_size/group_info[group_ind].r_max*1000))
           group_info[group_ind].x_send=feedback_header.x_r;
        else
        {  group_info[group_ind].target_rate=feedback_header.x_r;
           /* group_info[group_ind].increase_rate=
                              8*mean_packet_size/(group_info[group_ind].r_max^2); */
           /* the following is for the test harness, because test_harness can't give
            * accurate received_time
            */
            /*group_info[group_ind].start_point=feedback_header.received_time;*/
           group_info[group_ind].start_point=get_timestamp();
        }
      }
      /* the current CLR is not NULL, and the new feedback is not from the CLR.
       * Draft 3.3-2
       */
      else if (strcmp(group_info[group_ind].clr.receiver_id, 
                                 feedback_header.receiver_id)!=0)
      {  if(feedback_header.x_r < group_info[group_ind].x_send &&
                                 feedback_header.receiver_leave_flag==0)
         {  strcpy(group_info[group_ind].clr.receiver_id, feedback_header.receiver_id);
            group_info[group_ind].clr.receiver_timestamp=
                                 feedback_header.receiver_timestamp;
            /* the following line is for test harness, because it can't provide 
             * accurate received time
             */
            /*group_info[group_ind].clr.received_time=feedback_header.received_time; */
            group_info[group_ind].clr.received_time=get_timestamp();
            group_info[group_ind].clr.receiver_leave_flag=
                                 feedback_header.receiver_leave_flag;
            group_info[group_ind].clr_new_flag=1;
            group_info[group_ind].clr_non_feedback=0;
            group_info[group_ind].clr_leave_flag =feedback_header.receiver_leave_flag;
            group_info[group_ind].x_send=feedback_header.x_r;
            group_info[group_ind].target_rate=-1;
            group_info[group_ind].feedback_rate=-1;
         }
         /* Draft 3.3-3 */
         else if(group_info[group_ind].clr_leave_flag==1)
         {  /* At this point, the X_r in feedback may be either larger than x_send, 
             * or smaller than x_send. However, the draft does not specify what to do 
             * with the difference!!
             * Wether to choose a receiver with ( X_r < x_send && receiver_leave==1)
             * or else choose a receiver with ( X_r >x_send && receiver_leave==0)??
             */

            /* note: if the feedback_rate > x_send +8*mean_packet_size/r_max, 
             * will it finally increase gradually or not?
             * the draft does not specify that!! 
             * in the current implementation, just ignor if the new receiver is
             * going to leave.
             */
 
            if(feedback_header.x_r<group_info[group_ind].x_send)
            {  strcpy(group_info[group_ind].clr.receiver_id,
                                       feedback_header.receiver_id);
               group_info[group_ind].clr.receiver_timestamp=
                                       feedback_header.receiver_timestamp;
               /* note: the following line is for test hareness, because
                * it can't provide accurate received_time
                */
               /* group_info[group_ind].clr.received_time=
                                       feedback_header.received_time; */
               group_info[group_ind].clr.received_time=get_timestamp();
               group_info[group_ind].clr.receiver_leave_flag=
                                       feedback_header.receiver_leave_flag;

               group_info[group_ind].clr_new_flag=1;
               group_info[group_ind].clr_non_feedback=0;
               /* group_info[group_ind].clr_leave_flag =0; */
               group_info[group_ind].x_send=feedback_header.x_r;
               group_info[group_ind].target_rate=-1;
               /* note, the start_point may also need to be reset here. */
               group_info[group_ind].feedback_rate=-1;
            }
            else if(group_info[group_ind].feedback_rate<0 ||
                           feedback_header.x_r <group_info[group_ind].feedback_rate)
            {  strcpy(group_info[group_ind].clr.receiver_id, 
                                       feedback_header.receiver_id);
               group_info[group_ind].clr.receiver_timestamp=
                                       feedback_header.receiver_timestamp;
               /* note: the following line is for test harness, because it can't
                * provide accurate received_time
                */
               /*group_info[group_ind].clr.received_time=
                                       feedback_header.received_time; */
               group_info[group_ind].clr.received_time=get_timestamp();
               group_info[group_ind].clr.receiver_leave_flag=
                                       feedback_header.receiver_leave_flag;

               group_info[group_ind].clr_new_flag=1;
               group_info[group_ind].clr_non_feedback=0;
               /* group_info[group_ind].clr_leave_flag= 
                                       feedback_header.receiver_leave_flag;*/
               /* note, the clr_leave_flag should be set according to the
                * CLR.receiver_leave_flag at the end of each feedback round.
                */
               group_info[group_ind].feedback_rate=feedback_header.x_r;
            }
         }
      }
      /* the feedback is from the current CLR */
      else
      {  /* note: the draft does not specify the following case:
          *  1. CLR is NULL, and a new reciver send back a feedback, 
          *     says, with X_r=1000
          *  2. the x_send will increase gradually.
          *  3. Another feedback from CLR arrived with X_r=1200, 
          *     when x_send increased to 500
          *  4. the x_send will be set to x_send+8S/Rmax, and stop increase
          */
         /*strcpy(group_info[group_ind].clr.receiver_id, feedback_header.receiver_id);*/
         group_info[group_ind].clr.receiver_timestamp=
                                       feedback_header.receiver_timestamp;
         /* note: the following line is for test harness, because it can't provide
          * accurate received_Time 
          */
         /* group_info[group_ind].clr.received_time=feedback_header.received_time; */
         group_info[group_ind].clr.received_time=get_timestamp();
         group_info[group_ind].clr.receiver_leave_flag= 
                                       feedback_header.receiver_leave_flag;
         /*group_info[group_ind].clr_new_flag=1; */
         group_info[group_ind].clr_non_feedback=0;
         group_info[group_ind].clr_leave_flag = feedback_header.receiver_leave_flag;
         group_info[group_ind].target_rate=-1;
         /* note, the start_point may also need to be reset here. */
         group_info[group_ind].feedback_rate=-1;
         group_info[group_ind].x_send=min_value(feedback_header.x_r, 
               group_info[group_ind].x_send+8*group_info[group_ind].mean_packet_size/
               group_info[group_ind].r_max*1000);

      }


      /* note: non_clr feedbck should be insterted into the link list
       * otherwise save in group_info[group_ind].clr
       */

#ifdef UNIX
      pthread_mutex_unlock(&(group_info[group_ind].mutex));
#else
      ReleaseMutex(group_info[group_ind].mutex);
#endif
      return 0;

   }

   /* at this point, the feedback is not for the local sender */
   /****************************************************/
   /*   update minimum_x_supp of the targeted sender   */
   /****************************************************/
#ifdef UNIX
   pthread_mutex_unlock(&(group_info[group_ind].mutex));
#else
   ReleaseMutex(group_info[group_ind].mutex);
#endif
   /* update the mimum x_r from all the receivers
    * note, should wait for some mutex here.
    * search for the index of the sender_id,
    * if does not exist, create one;
    */
#ifdef UNIX
   pthread_mutex_lock(&(group_info[group_ind].other_senders_mutex));
#else
   WaitForSingleObject(group_info[group_ind].other_senders_mutex, INFINITE);
#endif
   for(i=0; i<20; i=i+1)
   {  if(strcmp(group_info[group_ind].other_senders[i].sender_list.sender_id, feedback_header.sender_id)==0)
      {  index=i;
         break;
      }
   }
   if(i>=20) /* a new sender, no previous entry found */
   {  for(i=0; i<20; i=i+1)
      {  if(strlen(group_info[group_ind].other_senders[i].sender_list.sender_id)==0)
         {  /* found a empty slot. */
            index=i;
            strcpy(group_info[group_ind].other_senders[i].sender_list.sender_id, feedback_header.sender_id);
            break;
         }
      }
   }

   if(i>=20) /* no empty slot*/
   {  printf("no empty slot for tfmcc receiver to create sender record\n\a\a");
      exit(-1);
   }

   if(strcmp(feedback_header.receiver_id, group_info[group_ind].other_senders[index].clr_id)!=0)
   {  if(feedback_header.x_r<group_info[group_ind].other_senders[index].minimum_x_supp)
         group_info[group_ind].other_senders[index].minimum_x_supp = feedback_header.x_r;
   }
#ifdef UNIX
   pthread_mutex_unlock(&(group_info[group_ind].other_senders_mutex));
#else
   ReleaseMutex(group_info[group_ind].other_senders_mutex);
#endif

   return 0;
}

/* note: what if the sender did not receive any feedback during a feedback round?
 *      will the r_max still be reduced to 0.9*R_max?
 *      If there is no any receiver at all, this will lead to the sender over active!
 */


/**************************************************************************************
 * For each multicast group there should be a thread_fb_nr.                           *
 * This thread is mainly for control the feedback round.                              *
 * The thread should be created by SRMP by calling search_tfmcc_mca()                 *
 * A feedback round can be dived into 2 T period, T=4*r_max                           *
 * During the first T period, if non-CLR feedback was received, current feedback      *
 *  should be terminated after the fisrt T period expired. Else, the feedback round   *
 *  keeps on for the second T period. During this period, whenever a non-CLR feedback *
 *  was received, a new feedback round should be started. Otherwise, the current      *
 *  feedback round should be terminated after the second T period expired             *
 *  1. wait T                                                                         *
 *   2. at the end of T, if non-CLR feedback received, end current feedback round     * 
 *   3. else, wait another T. this period can be terminated by event group_inf[].event*
 *   4. at the end of the second T, the current feedback round should be terminated   *
 *  5. at the end of T, update fb_nr, r_max, x_send, x_target                         *
 **************************************************************************************/
#ifdef UNIX
void * thread_fb_nr(void * mp)
#else
void thread_fb_nr(void * mp)
#endif
{  int i;
   unsigned short int r_max;
   double rate;
   unsigned short int current_time;
   int group_ind;
   double temp3;
#ifdef UNIX
   struct timespec delay_time;
   time_t current_systime;
#else
   SYSTEMTIME current_systime;
#endif

   group_ind=*(int*)mp;

#ifdef UNIX
   pthread_mutex_lock(&(group_info[group_ind].mutex));
#else
   WaitForSingleObject(group_info[group_ind].mutex, INFINITE);
#endif
   while(1)
   {  /*printf("\n   ********* a new feedback round rmax=%d ********\n", 
                  group_info[group_ind].r_max); */
      /*printf(".");
      fflush(stdout);*/
      r_max=group_info[group_ind].r_max;
      group_info[group_ind].x_supp=255e255;

      /* note: during the slow start stage:
       * at the start of the feedback round, set the
       *  group_info[group_ind].slowstart_rate to -1;
       * at the end of the feedback round, set the  
       *  group_info[group_ind].x_send=group_info[group_ind].slowstart_rate;
       */
      group_info[group_ind].slowstart_rate=-1;

      /* note: during the feedback round of a not slow start stage:
       * at the start of the feedback round, set the
       *    group_info[group_ind].feedback_rate to 0;
       * at the end of the feedback round, if the group_info[group_ind].feedback_rate
       *    is not 0, set group_info[group_ind].x_send=
       *    group_info[group_ind].feedback_rate
       * at the start of the feedback round
       *    note to check clr_non_feedback;
       *    note to check non_feedback;
       */

      /* T=4*R_max */
      /*printf("          point 1000 in thread_fb_nr\n"); */
      /***************/
      /*   First T   */
      /***************/
      for(i=0;i<4; i++)
      {
#ifdef UNIX
         pthread_mutex_unlock(&(group_info[group_ind].mutex));
         usleep(r_max*1000);
         pthread_cond_signal(&(group_info[group_ind].fbnr_event));
         pthread_mutex_lock(&(group_info[group_ind].mutex));
#else
         ReleaseMutex(group_info[group_ind].mutex);
         Sleep(r_max);
         SetEvent(group_info[group_ind].fbnr_event);   /* inform SRMP */
         /*ReleaseSemaphore(hSemfb_nr,1,&lcount);*/
         /*printf("point 2100 in thread_fb_nr\n"); */
         WaitForSingleObject(group_info[group_ind].mutex, INFINITE);
#endif
         /* have not received feedback from CLR for 10 RTT*/
         if(group_info[group_ind].clr_non_feedback>=10) 
         {  /* CLR is depicted as leaving */
            group_info[group_ind].clr.receiver_leave_flag=1;
            group_info[group_ind].clr_leave_flag=1;
            group_info[group_ind].clr_new_flag=0;
            group_info[group_ind].clr_non_feedback=0;
         }
         else
         {  group_info[group_ind].clr_non_feedback=
                                 group_info[group_ind].clr_non_feedback+1;
         }
         
         /* have not received feedback in 10 RTT */
         if(group_info[group_ind].non_feedback>=10)
         {  if(group_info[group_ind].x_send>=0)  /* readue the sending rate */
            {  group_info[group_ind].x_send=group_info[group_ind].x_send/2;
               if(group_info[group_ind].mean_packet_size<0)
               {  if(group_info[group_ind].x_send<DEFAULT_MEAN_PKT_SIZE/64)
                  {  group_info[group_ind].x_send=-0.5;
                        /* -0.5 means 1 packet per 64 seconds */
#ifdef UNIX
                     time(&current_systime);
                     group_info[group_ind].last_packet_time=current_systime;
#else
                     GetSystemTime(&current_systime);
                     group_info[group_ind].last_packet_time=
                        ((current_systime.wDay*24+current_systime.wHour)*60
                         +current_systime.wMinute)*60+current_systime.wSecond;
#endif
                  }
               }
               else
               {  if(group_info[group_ind].x_send<
                                 group_info[group_ind].mean_packet_size/64)
                  {  group_info[group_ind].x_send=-0.5;
                          /* -0.5 means 1 packet per 64 seconds */
#ifdef UNIX
                     time(&current_systime);
                     group_info[group_ind].last_packet_time=current_systime;
#else
                     GetSystemTime(&current_systime);
                     group_info[group_ind].last_packet_time=
                        ((current_systime.wDay*24+current_systime.wHour)*60
                         +current_systime.wMinute)*60+current_systime.wSecond;
#endif
                  }
               }
            }

            group_info[group_ind].r_max=INITIAL_RMAX;
            /* printf("reset r_max\n"); */
            group_info[group_ind].non_feedback=0;
         }
         else
         {  group_info[group_ind].non_feedback=group_info[group_ind].non_feedback+1;
         }
      }
      /* printf("          point 3000 in thread_fb_nr\n"); */

      /* after time T, check to see if Non_CLR feedback has been received
       *   if yes, the current feedback round will terminate.
       */
      /**************************************************************/
      /* end of the first T, update fb_nr, r_max, x_send, x_target  */
      /**************************************************************/
      if( group_info[group_ind].feedback_received_flag==1)
      {  group_info[group_ind].feedback_received_flag=0;
         group_info[group_ind].clr_leave_flag=
                                 group_info[group_ind].clr.receiver_leave_flag;
         group_info[group_ind].fb_nr=(group_info[group_ind].fb_nr+1)%16;
         /* check to see if any feedback with a larger RTT has been received*/
         if (group_info[group_ind].no_large_rmax==1)
         {  if(group_info[group_ind].mean_packet_size<0)
            {  group_info[group_ind].r_max=
                     (unsigned short int)max_value(0.9 * group_info[group_ind].r_max,
                              DEFAULT_MEAN_PKT_SIZE/group_info[group_ind].x_send);
            }
            else
               group_info[group_ind].r_max=
                     (unsigned short int)max_value(0.9 * group_info[group_ind].r_max,
                  group_info[group_ind].mean_packet_size/group_info[group_ind].x_send);
            /* note: here is a problem with the tfmcc draft
             *   the sender may cut its sending rate to half, when there is no feedback
             *      during 10*RTT
             *   this kind of decrease sending rate does not mean a longer rtt is 
             *     suitable
             *   so I just limit the increasement of RTT to maximum INITIAL
             */
            if(group_info[group_ind].r_max>=INITIAL_RMAX)
               group_info[group_ind].r_max=INITIAL_RMAX;
         }
         else
            group_info[group_ind].no_large_rmax=1;

         /*  set the sending rate */
         if(group_info[group_ind].x_send < group_info[group_ind].target_rate)
         {  if(group_info[group_ind].x_send<0)
               group_info[group_ind].x_send=0;
            current_time=get_timestamp();
            if(group_info[group_ind].in_slowstart_flag==1)
            {  rate=group_info[group_ind].x_send+
                     (current_time-group_info[group_ind].start_point)
                     *group_info[group_ind].increase_rate;
               printf("      increasing rate=%f\n", 
                     group_info[group_ind].increase_rate);
            }
            else
            {  if(group_info[group_ind].mean_packet_size<0)
               {  rate=group_info[group_ind].x_send+
                     (current_time-group_info[group_ind].start_point)*
                     8*DEFAULT_MEAN_PKT_SIZE/group_info[group_ind].r_max/
                     group_info[group_ind].r_max*1000;
               }
               else
               {  rate=group_info[group_ind].x_send+
                     (current_time-group_info[group_ind].start_point)*8*
                     group_info[group_ind].mean_packet_size/
                     group_info[group_ind].r_max/group_info[group_ind].r_max*1000;
               }
            }

            group_info[group_ind].start_point=current_time;

            if(rate>=group_info[group_ind].target_rate)
            {  rate=group_info[group_ind].target_rate;
               group_info[group_ind].increase_rate=0;
               group_info[group_ind].target_rate=-1;
               group_info[group_ind].start_point=65535;
            }

            group_info[group_ind].x_send=rate;
         }


         /* update the new targe_rate */
         if(group_info[group_ind].in_slowstart_flag==0)
         {  if(group_info[group_ind].feedback_rate>0)
            {  group_info[group_ind].target_rate=group_info[group_ind].feedback_rate;
               group_info[group_ind].feedback_rate=-1;
               group_info[group_ind].start_point=get_timestamp();
            }
         }
         /* update target rate and CLR in slow start stage */
         else
         {  if(group_info[group_ind].slowstart_rate>0)
            {  if(group_info[group_ind].slowstart_rate>group_info[group_ind].x_send)
               {  group_info[group_ind].target_rate=
                                             group_info[group_ind].slowstart_rate;
                  group_info[group_ind].increase_rate=
                    (group_info[group_ind].target_rate-group_info[group_ind].x_send)
                    /group_info[group_ind].r_max;
                  group_info[group_ind].start_point=get_timestamp();
               }
               else
               {  group_info[group_ind].target_rate=-1;
                  group_info[group_ind].increase_rate=-1;
                  group_info[group_ind].x_send=group_info[group_ind].slowstart_rate;
               }

               group_info[group_ind].slowstart_rate=-1;

               /* note to check if target_clr is NULL(all zero) */
               if(strcmp(group_info[group_ind].clr.receiver_id,
                  group_info[group_ind].target_clr.receiver_id)!=0
                   && strlen(group_info[group_ind].target_clr.receiver_id)>0)
               {  group_info[group_ind].clr_new_flag=1;
                  memcpy(&group_info[group_ind].clr, &group_info[group_ind].target_clr,
                                          sizeof(struct feedback_list_t));
                  group_info[group_ind].clr_leave_flag=
                                       group_info[group_ind].clr.receiver_leave_flag;
                  group_info[group_ind].clr_non_feedback=0;

                  memset(&group_info[group_ind].target_clr, 0, 
                                                   sizeof(struct feedback_list_t));
               }

            }
         }

         /* terminate the current fbnr */
         /* ReleaseSemaphore(hSemfb_nr,1,&lcount);  */
         /* this is to inform SRMP that a new feedback round has start */
         /* printf("          point 5000 in thread_fb_nr\n"); */
         continue;

      }

      /**********************/
      /*   the second T     */
      /**********************/
      /* else, Non-CLR feedback was not received during the first T */
      /* printf("   +++++++++++ first 4 RTTs expired ++++++++++\n");*/
      /* ResetEvent(group_info[group_ind].event);*/
      for(i=0;i<4;i++)
      {  /*Sleep(r_max);*/

#ifdef UNIX
         memset(&delay_time, 0, sizeof(struct timespec));
         delay_time.tv_sec=r_max/1000;
         delay_time.tv_nsec=(r_max%1000)*1000000;
         pthread_cond_reltimedwait_np(&(group_info[group_ind].event),
                                    &(group_info[group_ind].mutex), &delay_time);
         pthread_cond_signal(&(group_info[group_ind].fbnr_event));
#else
         ReleaseMutex(group_info[group_ind].mutex);
         WaitForSingleObject(group_info[group_ind].event, r_max );
         WaitForSingleObject(group_info[group_ind].mutex, INFINITE);
         /* to inform the SRMP a new RMAX, in case of sending rate is 1 segment/Rmax */
         SetEvent(group_info[group_ind].fbnr_event);
#endif

         /*ReleaseSemaphore(hSemfb_nr,1,&lcount);*/
         /*printf(" point 6000 in thread fb_nr\n");*/
         if(group_info[group_ind].clr_non_feedback>=10)
         {  group_info[group_ind].clr_leave_flag=1;
            group_info[group_ind].clr.receiver_leave_flag=1;
            group_info[group_ind].clr_new_flag=0;
            group_info[group_ind].clr_non_feedback=0;
         }
         else
         {  group_info[group_ind].clr_non_feedback=
                                       group_info[group_ind].clr_non_feedback+1;
         }

         if(group_info[group_ind].non_feedback>=10)
         {  if(group_info[group_ind].x_send>0)
            {  group_info[group_ind].x_send=group_info[group_ind].x_send/2;

               if(group_info[group_ind].mean_packet_size<0)
                  temp3=DEFAULT_MEAN_PKT_SIZE;
               else
                  temp3=group_info[group_ind].mean_packet_size;

               if(group_info[group_ind].x_send<temp3/64)
                  group_info[group_ind].x_send=temp3/64;
            }
            /* note, there is no such specification in the draft */
            group_info[group_ind].r_max=INITIAL_RMAX;
            /*printf("reset r_max\n"); */
            group_info[group_ind].non_feedback=0;
         }
         else
         {  group_info[group_ind].non_feedback=group_info[group_ind].non_feedback+1;
         }
         /* terminate by receiving a Non_CLR feedback */
         if( group_info[group_ind].feedback_received_flag==1)
         {  /*printf("++++ point 3600\n");*/
            break;
         }


      }
      /***************************************************************/
      /* end of the second T, update fb_nr, r_max, x_send, x_target  */
      /***************************************************************/

      group_info[group_ind].feedback_received_flag=0;
      group_info[group_ind].clr_leave_flag=
                                    group_info[group_ind].clr.receiver_leave_flag;
      /* note: may be wrong! */
      group_info[group_ind].fb_nr=(group_info[group_ind].fb_nr+1)-
         (int)(group_info[group_ind].fb_nr+1)/16*(group_info[group_ind].fb_nr+1);

      /* check to see if any feedback with a larger RTT has been received */
      if (group_info[group_ind].no_large_rmax==1)
      {  if(group_info[group_ind].mean_packet_size<0)
            group_info[group_ind].r_max=(unsigned short int)max_value(
                     0.9 * group_info[group_ind].r_max,
                     DEFAULT_MEAN_PKT_SIZE/group_info[group_ind].x_send);
         else
            group_info[group_ind].r_max=(unsigned short int)max_value(
                 0.9 * group_info[group_ind].r_max, 
                 group_info[group_ind].mean_packet_size/group_info[group_ind].x_send);
        /* note: here is a problem with the tfmcc draft
         *    the sender halve its sending rate and increase the r_max, when there is
         *    no feedback during 10*RTT
         *    this kind of decrease sending rate does not mean a longer rtt is 
         *    reasonable
         *    so I just limit the increasement of RTT to maximum INITIAL
         */
         if(group_info[group_ind].r_max>=INITIAL_RMAX)
            group_info[group_ind].r_max=INITIAL_RMAX;
      }
      else
         group_info[group_ind].no_large_rmax=1;

      /* set the sending rate */
      if(group_info[group_ind].x_send < group_info[group_ind].target_rate)
      {  if(group_info[group_ind].x_send<0)
            group_info[group_ind].x_send=0;
         current_time=get_timestamp();
         if(group_info[group_ind].in_slowstart_flag==1)
         {  rate=group_info[group_ind].x_send+
                  (current_time-group_info[group_ind].start_point)
                  *group_info[group_ind].increase_rate;
            /*printf("      increase rate=%f\n", group_info[group_ind].increase_rate);*/
         }
         else
         {  if(group_info[group_ind].mean_packet_size<0)
               rate=group_info[group_ind].x_send+
                  (current_time-group_info[group_ind].start_point)*8*
                  DEFAULT_MEAN_PKT_SIZE/group_info[group_ind].r_max/
                  group_info[group_ind].r_max*1000;
            else
               rate=group_info[group_ind].x_send+
                  (current_time-group_info[group_ind].start_point)*8*
                  group_info[group_ind].mean_packet_size/group_info[group_ind].r_max/
                  group_info[group_ind].r_max*1000;
         }

         group_info[group_ind].start_point=current_time;

         if(rate>=group_info[group_ind].target_rate)
         {  rate=group_info[group_ind].target_rate;
            group_info[group_ind].increase_rate=0;
            group_info[group_ind].target_rate=-1;
            group_info[group_ind].start_point=65535;
         }
         group_info[group_ind].x_send=rate;
      }

      /* update the new target rate */
      if(group_info[group_ind].in_slowstart_flag==0)
      {  if(group_info[group_ind].feedback_rate>0)
         {  group_info[group_ind].target_rate=group_info[group_ind].feedback_rate;
            group_info[group_ind].feedback_rate=-1;
            group_info[group_ind].start_point=get_timestamp();
         }
      }
      /* update target rate and CLR in slow start stage */
      else
      {  if(group_info[group_ind].slowstart_rate>0)
         {  if(group_info[group_ind].slowstart_rate>group_info[group_ind].x_send)
            {  group_info[group_ind].target_rate=group_info[group_ind].slowstart_rate;
               group_info[group_ind].increase_rate=
                  (group_info[group_ind].target_rate-group_info[group_ind].x_send)/
                  group_info[group_ind].r_max;
               group_info[group_ind].start_point=get_timestamp();
            }
            else
            {  group_info[group_ind].target_rate=-1;
               group_info[group_ind].increase_rate=-1;
               group_info[group_ind].x_send=group_info[group_ind].slowstart_rate;
            }
            group_info[group_ind].slowstart_rate=-1;

            /* note to check if target_clr is NULL(all zero) */
            if(strcmp(group_info[group_ind].clr.receiver_id, 
                        group_info[group_ind].target_clr.receiver_id)!=0)
            {  group_info[group_ind].clr_new_flag=1;
            }
            memcpy(&group_info[group_ind].clr, &group_info[group_ind].target_clr,
                                                   sizeof(struct feedback_list_t));
            group_info[group_ind].clr_leave_flag=
                                       group_info[group_ind].clr.receiver_leave_flag;
            group_info[group_ind].clr_non_feedback=0;

            /*memset(group_info[group_ind].target_clr, 0, 20); */

         }
      }

      /*ReleaseSemaphore(hSemfb_nr,1,&lcount);*/

      /*printf("          point 8000 in thread_fb_nr\n");*/

   }  /* end of while loop */
#ifdef UNIX
   pthread_mutex_unlock(&(group_info[group_ind].mutex));
#else
   ReleaseMutex(group_info[group_ind].mutex);
#endif
} /* end of function */



/**************************************************************************************/
/* This function should be called by SRMP, when each time a bundle was received       */
/*   0. if a new feedback round started, clear up old record and feedbacks            */
/*   1. check if the bundle_sn is a new one, if NOT, discard it.                      */
/*   2. if the receiver_id in the bundle_header equals to local receiver              */
/*       then, update r_max,RTT, have_rtt_flag                                        */
/*      else, update r_max                                                            */
/*   3. update packet_history, calculate packet loss event rate                       */
/*   4. update the is_clr_flag                                                        */
/*   5. if local receiver became a new CLR, schedule a feedback                       */
/*      else, if a new feedback round started, remove all old feedback timer,         */
/*               and sechdule a new one                                               */
/**************************************************************************************/
int upon_receive_bundle(struct bundle_header_t bundle_header,
                        double packet_size, unsigned long int mca  )
{  int i;
   int j;
   int index;
   char sender_id[20];
   int new_fbnr; 
   unsigned short int r_sample;
   int old_is_clr_flag;
   int old_have_rtt_flag;
   double rtt_ratio;
   int s_before;
   int s_after;
   unsigned short int t_before;
   unsigned short int t_after;
   int delay;   /* the feedback backoff time, as described by tfmcc draft 4.5 */


   int unwrap_sn; /* a temporary variable to calculate unrapped_sn */
   int temp_sn;
   struct time_event_t * new_event;
   struct time_event_t * time_search;
   int group_ind;

   new_fbnr=0;
   group_ind=search_tfmcc_mca(mca, 0);
   if (group_ind<0)
   {  return(-1);
   }
   if(r_time_axis==-1)
   {  r_time_axis=0;
      r_start=get_timestamp();
   }

   /* because in SRMP added in packet loss thread, so only SRMP can 
    * calculate home many packet was actually received 
    */
   /* number_packet_received++;*/
   r_stop=get_timestamp();
   r_delay=r_stop - r_start;
   if(r_delay > 5000)
   {  packet_loss_ratio=(double)number_packet_lost/number_packet_received;
      fprintf(LossRecord, "%d\t%f\n", r_time_axis, packet_loss_ratio);
      r_time_axis=r_time_axis + r_delay;
      r_start=r_stop;
      number_packet_received=0;
      number_packet_lost=0;
   }


   /*   printf("\n-------------- received new bundle\n");*/
#ifdef UNIX
   pthread_mutex_lock(&(group_info[group_ind].other_senders_mutex));
#else
   WaitForSingleObject(group_info[group_ind].other_senders_mutex, INFINITE);
#endif
   /*   current_time=get_timestamp();*/
   /*****************************************/
   /* search for the index of the sender_id */
   /*****************************************/
   /* if does not exist, create one; */
   strcpy(sender_id, bundle_header.sender_id);
   for(i=0; i<20; i=i+1)
   {  if(strcmp(group_info[group_ind].other_senders[i].sender_list.sender_id,
               sender_id)==0)
      {  index=i;
         break;
      }
   }

   if(i>=20) /* a new sender, no previous entry found */
   {  if(group_info[group_ind].current_receiver_leave==1)
      {  /* the receiver is in leaving stage */
         printf("   receiver leaving, the feedback is ignored\n");
         return 0;
      }

      for(i=0; i<20; i=i+1)
      {  if(strlen(group_info[group_ind].other_senders[i].sender_list.sender_id)==0)
         {  /* found a empty slot. Note to reset when sender left*/
            index=i;
            break;
         }
      }
   }
   if(i>=20) /* no empty slot */
   {  printf("no empty slot for tfmcc receiver to create sender record\n");
      exit(-1);
   }

   /*show_bundle_header(&bundle_header);*/
   /***************************/
   /* update mean packet size */
   /***************************/
   if(group_info[group_ind].other_senders[index].mean_packet_size<0)
   {  group_info[group_ind].other_senders[index].mean_packet_size=packet_size;
   }
   else
   {  group_info[group_ind].other_senders[index].mean_packet_size=
            group_info[group_ind].other_senders[index].mean_packet_size+packet_size;
      group_info[group_ind].other_senders[index].mean_packet_size=
                     group_info[group_ind].other_senders[index].mean_packet_size /2;
   }



   /* check to see if a new feedback round started*/
   if(bundle_header.fb_nr > 
      group_info[group_ind].other_senders[index].sender_list.fb_nr ||
      (bundle_header.fb_nr+8) <= 
      group_info[group_ind].other_senders[index].sender_list.fb_nr)
   {  /********************************/
      /* new feedback round started   */
      /********************************/
      new_fbnr=1;  /* note, should schedule a timer */
      /* reset this value, this value should be set in feedback_received().*/
      group_info[group_ind].other_senders[index].minimum_x_supp=1e20;

      /* receive is leaving, and the index to the sender still exists
       * this means the receiver is the CLR for current sender ( or ever been the CLR)
       * destroy data structure for the sender
       */
      if(group_info[group_ind].current_receiver_leave==1)
      {  /* first remove all the record in group_info[group_ind].time_event_head,
          * that destined to the indexed sender 
          */
#ifdef UNIX
         pthread_mutex_lock(&(group_info[group_ind].time_event_mutex));
#else
         WaitForSingleObject(group_info[group_ind].time_event_mutex, INFINITE);
#endif

         if(group_info[group_ind].time_event_head!=NULL)
         {  struct time_event_t * temp_time_event;
            struct time_event_t * temp2_time_event;

            temp_time_event=NULL;
            temp2_time_event=NULL;

            /* the first time event */
            if (group_info[group_ind].time_event_head->index == index)
               group_info[group_ind].cancel_flag=1;
            /* move the link list head to the first record not destined to the 
             * indexed sender
             */
            temp_time_event=group_info[group_ind].time_event_head;
            while( (temp_time_event!=NULL) && (temp_time_event->index==index) )
            {  group_info[group_ind].time_event_head=temp_time_event->next;
               free(temp_time_event);
               temp_time_event=group_info[group_ind].time_event_head;
            }
            if(temp_time_event==NULL)
               group_info[group_ind].time_event_tail=NULL;

            /* then remove all the remainning records that destined to the 
             * indexed sender
             */
            if(temp_time_event!=NULL)
               temp2_time_event=temp_time_event->next;
            while(temp_time_event!=NULL)
            {  if((temp2_time_event!=NULL) && (temp2_time_event->index==index) )
               {  temp2_time_event=temp2_time_event->next;
                  free(temp_time_event->next);
                  temp_time_event->next=temp2_time_event;
               }
               else if(temp2_time_event!=NULL)
               {  temp2_time_event=temp2_time_event->next;
                  temp_time_event=temp_time_event->next;
               }
               else
               {  group_info[group_ind].time_event_tail=temp_time_event;
                  break;
               }
            }
         }
#ifdef UNIX
         pthread_mutex_unlock(&(group_info[group_ind].time_event_mutex));
#else
         ReleaseMutex(group_info[group_ind].time_event_mutex);
#endif

         /* second clear up receiver_info for the indexed sender */
         memset(&group_info[group_ind].other_senders[index], 0, 
                                    sizeof(struct other_sender_t));
         group_info[group_ind].other_senders[index].s_old=-1;
         group_info[group_ind].other_senders[index].have_loss_flag=0;
         group_info[group_ind].other_senders[index].minimum_x_supp=1e20;
         strcpy(group_info[group_ind].other_senders[index].receiver_id, ip_address);
         group_info[group_ind].other_senders[index].sender_list.fb_nr=-1;
         group_info[group_ind].other_senders[index].mean_packet_size=-1;
         group_info[group_ind].other_senders[index].number_of_loss=0;

         for(i=0; i<20; i=i+1)
         {  group_info[group_ind].other_senders[index].packet_history[i].sn=-1;
            group_info[group_ind].other_senders[index].packet_history[i].received_time=
                                                                                 65535;
         }

         /* third, check to see if no sender to wait. If yes, send signal to inform 
          * safely leave
          */
         for(i=0; i<20; i=i+1)
         {  if(group_info[group_ind].other_senders[i].sender_list.fb_nr>=0)
               break;
         }
         if(i>=20)
         {  printf("\n\n   *** the receiver can leave safely ***\a\a\a\n\n\n");
            /* note: should send a signal to inform SRMP here*/
         }

         return 0;
      }

   }


   old_is_clr_flag=group_info[group_ind].other_senders[index].is_clr_flag;
   old_have_rtt_flag=
               group_info[group_ind].other_senders[index].feedback_list.have_rtt_flag;

   /******************************/
   /*  update bundle sn          */
   /******************************/
   /* if it is a new bundle, sn > current sn, then copy it, else discard it
    * note to set the sn value when sending out bundle
    * note to take care of the wrappecd sn number
    * note: this means if a packet was late because of jitter, the packet
    *       will be discarded.
   */
   if(bundle_header.sn>= group_info[group_ind].other_senders[index].sender_list.sn ||
      (bundle_header.sn+32768) <= 
               group_info[group_ind].other_senders[index].sender_list.sn)
   {  memcpy(&group_info[group_ind].other_senders[index].sender_list, 
                              &bundle_header, sizeof(struct bundle_header_t));
   }
   else
   {
#ifdef UNIX
      pthread_mutex_unlock(&(group_info[group_ind].other_senders_mutex));
#else
      ReleaseMutex(group_info[group_ind].other_senders_mutex);
#endif
      printf("error sn received!\a\n");
      return 0;
   }


   /* update rtt if the receiver_id in the bundle received equeals to that of 
    * current receiver 
    */
   if(strcmp(group_info[group_ind].other_senders[index].receiver_id, 
                                                bundle_header.receiver_id)==0)
   {  /****************************************************/
      /* the bundle has the receiver_id as local receiver */
      /****************************************************/
      if( (bundle_header.received_time < bundle_header.receiver_timestamp)
         && ((bundle_header.receiver_timestamp - bundle_header.received_time) <32768)
         )
      {  r_sample=0+SIMU_DELAY;
      }
      else
      {  r_sample= bundle_header.received_time - bundle_header.receiver_timestamp
                     +SIMU_DELAY;
      }


      if(old_have_rtt_flag==0)
      {  group_info[group_ind].other_senders[index].feedback_list.have_rtt_flag=1;
         group_info[group_ind].other_senders[index].feedback_list.rtt=r_sample;
      }
      else
      {  if(bundle_header.is_clr_flag==1)
            rtt_ratio=Q_HIS_CLR;
         else
            rtt_ratio=Q_HIS;

         group_info[group_ind].other_senders[index].feedback_list.rtt = 
             (unsigned short int)(((1-rtt_ratio)*r_sample +
             rtt_ratio*group_info[group_ind].other_senders[index].feedback_list.rtt) );
      }
      /* printf("   RTT=%d   ",
                  group_info[group_ind].other_senders[index].feedback_list.rtt); */
   }
   else /* if the bundle is not for the current receiver, then if the receiver has not
         * measured its rtt, then record the r_max as its rtt
         */
   {  if(group_info[group_ind].other_senders[index].feedback_list.have_rtt_flag==0)
      {  group_info[group_ind].other_senders[index].feedback_list.rtt=
                                                               bundle_header.r_max;
      }

   }

      if(group_info[group_ind].other_senders[index].feedback_list.rtt==0)
         group_info[group_ind].other_senders[index].feedback_list.rtt=1;


   /******************************/
   /*  update packet loss event  */
   /******************************/
   /*    s_loss is the sn of the packet which is the most recent confirmed to be lost
    *    s_old and t_old is the packet that start the most recent loss event
    *    insert the new packet to packet_history;
    */
   unsigned short delay2;
   unsigned short time_interval;
   int total_bundle_len;
   time_interval=0;
   total_bundle_len=0;
   for(i=0; i<19; i=i+1)
   {  group_info[group_ind].other_senders[index].packet_history[i].received_time=
          group_info[group_ind].other_senders[index].packet_history[i+1].received_time;
      group_info[group_ind].other_senders[index].packet_history[i].sn=
                     group_info[group_ind].other_senders[index].packet_history[i+1].sn;
      group_info[group_ind].other_senders[index].packet_history[i].len=
                    group_info[group_ind].other_senders[index].packet_history[i+1].len;
      if(group_info[group_ind].other_senders[index].packet_history[i].received_time ==
            65535)
         delay2=MEAN_RECV_PERIOD+1;
      else
         delay2=bundle_header.received_time-
            group_info[group_ind].other_senders[index].packet_history[i].received_time;
      if(delay2>MEAN_RECV_PERIOD)
      {  group_info[group_ind].other_senders[index].packet_history[i].len=0;
         time_interval=MEAN_RECV_PERIOD;
      }
      else
      {  total_bundle_len=total_bundle_len+
                     group_info[group_ind].other_senders[index].packet_history[i].len;
      }

   }
   group_info[group_ind].other_senders[index].packet_history[19].received_time=
                                                         bundle_header.received_time;
   group_info[group_ind].other_senders[index].packet_history[19].sn=
                                                                     bundle_header.sn;
   group_info[group_ind].other_senders[index].packet_history[19].len=
                                                                    (int) packet_size;
   total_bundle_len=(int) (total_bundle_len+packet_size);
/*
   for(i=0; i<20; i=i+1)
   {  printf("%d  ",
         group_info[group_ind].other_senders[index].packet_history[i].received_time);
   }
   printf("\n");
   for(i=0; i<20; i=i+1)
   {  printf("%d  ", 
            group_info[group_ind].other_senders[index].packet_history[i].len);
   }
   printf("\n");
*/

   if(time_interval==0)
      time_interval=bundle_header.received_time -
            group_info[group_ind].other_senders[index].packet_history[0].received_time;

   group_info[group_ind].other_senders[index].x_recv=
                                       (double)total_bundle_len*8/time_interval*1000;

   /*printf("x_recv=%f      ", group_info[group_ind].other_senders[index].x_recv);*/

   /* if no packet loss, then set s_loss to the sn of the oldest packet
    * in packet history
    * note to take into account SN wrapped
   */
   if(group_info[group_ind].other_senders[index].have_loss_flag==0)
   {  for(i=0; i<20; i=i+1)
      {  if(group_info[group_ind].other_senders[index].packet_history[i].sn>=0)
         {  group_info[group_ind].other_senders[index].s_loss=
                     group_info[group_ind].other_senders[index].packet_history[i].sn;
            break;
         }
      }
   }

   if((group_info[group_ind].other_senders[index].packet_history[19].sn+32768) <
               group_info[group_ind].other_senders[index].s_loss) /* sn wrapped */
   {  unwrap_sn=group_info[group_ind].other_senders[index].packet_history[19].sn+65536;
   }
   else
   {  unwrap_sn=group_info[group_ind].other_senders[index].packet_history[19].sn;
   }

/* printf("\n ------ s_loss=%d, unwrap_sn=%d\n", 
            group_info[group_ind].other_senders[index].s_loss, unwrap_sn); */
   for(i=group_info[group_ind].other_senders[index].s_loss; i<=(unwrap_sn - 3); i=i+1)
   {  /* find appropriate index for s_before */
      if(i<65536)
         temp_sn=i;
      else
         temp_sn=i-65536;

/*    for(j=0; j<5; j=j+1)
         printf("%d, ", 
                  group_info[group_ind].other_senders[index].packet_history[j].sn);
      printf("\n");
*/
      for(j=0; j<18; j=j+1)
      {  if((group_info[group_ind].other_senders[index].packet_history[j+1].sn>temp_sn)
            || ((group_info[group_ind].other_senders[index].packet_history[j+1].sn+
                  32768) < temp_sn))
            break; /* j is the index in packet_history[], point to s_before */

      }

      if(group_info[group_ind].other_senders[index].packet_history[j].sn==temp_sn)
      {  group_info[group_ind].other_senders[index].s_loss=temp_sn;
         continue;
      }

      number_packet_lost++;

      /* at this point, there must be some packet loss event
       *  packet_history[j].sn<i, and packet_history[j+1].sn>i
       *  this means that the packet with sn==i has been lost
       *  (sn+3 has already arrived, see the for loop)
       */
      group_info[group_ind].other_senders[index].have_loss_flag=1;
      group_info[group_ind].other_senders[index].s_loss=temp_sn;

      /* this is a extreme case: the receiver is not the CLR, so feedback may be send
       * out occasionally and the sender's sending rate is very high, the the packet
       * loss rate is high and since the RTT is a large value,  large amount of packet
       * loss belongs to the same loss event. so the result loss_event_interval is 
       * still large value. so the result x_r is a large value
       * This is some kind of dead lock.
       * Here I try to add in the following machenism: if the packet loss reached 
       * 5 ( or a ratio), then start a new loss interval, and send out feedback 
       * immediately
       * number_of_loss below is used to record how many packet lost, since the most
       * recent loss event
       */
      group_info[group_ind].other_senders[index].number_of_loss=
                        group_info[group_ind].other_senders[index].number_of_loss+1;

      s_before=group_info[group_ind].other_senders[index].packet_history[j].sn;
      s_after=group_info[group_ind].other_senders[index].packet_history[j+1].sn;
      t_before=
            group_info[group_ind].other_senders[index].packet_history[j].received_time;
      t_after=
          group_info[group_ind].other_senders[index].packet_history[j+1].received_time;
      if(s_before > s_after)  /* deal with sn wrap */
         s_after=s_after+65536;

      group_info[group_ind].other_senders[index].t_loss=
               t_before + (t_after - t_before)* (i- s_before)/(s_after-s_before);
      /* at this point (received_time=t_loss, sn=temp_sn) is current processed lost
       * packet
       * calculate inter-loss event Interval, tmfcc draft 5.3
       */

/*    printf("new loss event: t_old=%d, s_old=%d\n", 
            group_info[group_ind].other_senders[index].t_old, 
            group_info[group_ind].other_senders[index].s_old);
      printf("                t_loss=%d, temp_sn=%d\n", 
            group_info[group_ind].other_senders[index].t_loss, temp_sn);
      printf("                rtt=%d \n\n", 
            group_info[group_ind].other_senders[index].feedback_list.rtt);
*/

      if(group_info[group_ind].other_senders[index].s_old<0) /* the first loss event */
      {  group_info[group_ind].other_senders[index].t_old=
                                 group_info[group_ind].other_senders[index].t_loss;
         group_info[group_ind].other_senders[index].s_old=temp_sn;
      }
      else if((group_info[group_ind].other_senders[index].t_old+
                  group_info[group_ind].other_senders[index].feedback_list.rtt)
               <group_info[group_ind].other_senders[index].t_loss) /* a new loss event*/
      {  /* shift left loss_interval[10] */
         for(j=0; j<9; j=j+1)
            group_info[group_ind].other_senders[index].loss_interval[j]=
                        group_info[group_ind].other_senders[index].loss_interval[j+1];
         group_info[group_ind].other_senders[index].loss_interval[9]=
                  i-group_info[group_ind].other_senders[index].s_old;

         /* at this point, <t_loss,temp_sn> marks the start point of a new loss event */
         group_info[group_ind].other_senders[index].t_old=
                                 group_info[group_ind].other_senders[index].t_loss;
         group_info[group_ind].other_senders[index].s_old=temp_sn;
         group_info[group_ind].other_senders[index].number_of_loss=0;


      }
      else if (group_info[group_ind].other_senders[index].number_of_loss >
                                                               MAXIMUM_LOSS_PER_EVENT)
      {  for(j=0; j<9; j=j+1)
            group_info[group_ind].other_senders[index].loss_interval[j]=
                        group_info[group_ind].other_senders[index].loss_interval[j+1];
         group_info[group_ind].other_senders[index].loss_interval[9]=
                                    i-group_info[group_ind].other_senders[index].s_old;

         /* at this point, <t_loss,temp_sn> marks the start point of a new
          *   loss event
          */
         group_info[group_ind].other_senders[index].t_old=
                                 group_info[group_ind].other_senders[index].t_loss;
         group_info[group_ind].other_senders[index].s_old=temp_sn;
         group_info[group_ind].other_senders[index].number_of_loss=0;
#ifdef UNIX
         pthread_cond_signal(&(group_info[group_ind].time_event));
#else
         SetEvent(group_info[group_ind].time_event);
#endif

      }

   } /* end for loop */
   /*printf("\n");*/

/* double mean_loss_interval=0;
   double desired_rate;
   double p,q;
   if(group_info[group_ind].other_senders[index].have_loss_flag==1)
   {  for(i=0; i<8; i=i+1)
      {  mean_loss_interval=mean_loss_interval+
               weight[i]*group_info[group_ind].other_senders[index].loss_interval[i+2];
      }
      mean_loss_interval=mean_loss_interval/6;
      printf("point 2000\n");
      
      Initializing the loss history after the first loss event, as tfmcc draft 5.6
      if(mean_loss_interval == 0)
      {  group_info[group_ind].other_senders[index].loss_interval[9]=
            group_info[group_ind].other_senders[index].x_recv
            *group_info[group_ind].other_senders[index].feedback_list.rtt
            /1000/sqrt(2.0/3.0)
            /group_info[group_ind].other_senders[index].mean_packet_size;
         group_info[group_ind].other_senders[index].loss_interval[9]=
               pow(group_info[group_ind].other_senders[index].loss_interval[9], 2);
         mean_loss_interval=
               group_info[group_ind].other_senders[index].loss_interval[9]/6;
         printf("mean_loss_interval=%f\a\n", mean_loss_interval);
      }
   
      calculate the desired rate as TCP throughput equation
      p=1/mean_loss_interval;
      q=(sqrt(p*2/3)+p*12*sqrt(p*3/8)*(1+32*p*p));
      desired_rate=(double)8*
                  group_info[group_ind].other_senders[index].mean_packet_size/
                  group_info[group_ind].other_senders[index].feedback_list.rtt*1000;
      desired_rate=desired_rate/q;

   }
   else
   {  desired_rate=2*group_info[group_ind].other_senders[index].x_recv;
   }
   printf("p=%f, q=%f,  desired_rate=%f\n",p,q,  desired_rate);
*/
   /* end of update packet loss event */

   /* note: there are some problem with tfmcc draft 4.3.3. 
    * So 4.3.3 is not implemented here.
    */

   /* update lose event rate*/

   /************************/
   /*  update is_clr_flag  */
   /************************/
   if(strcmp(group_info[group_ind].other_senders[index].receiver_id,
                                                    bundle_header.receiver_id)==0)
   {  group_info[group_ind].other_senders[index].is_clr_flag =
                                                         bundle_header.is_clr_flag;
      if(bundle_header.is_clr_flag==0)
      {  strcpy(group_info[group_ind].other_senders[index].clr_id, 
                                                         bundle_header.receiver_id);
      }
   }
   else
   {  if(bundle_header.is_clr_flag==1)
      {  group_info[group_ind].other_senders[index].is_clr_flag=0;
         strcpy(group_info[group_ind].other_senders[index].clr_id, 
                                                         bundle_header.receiver_id);
      }
   }
   /************************************/
   /*  schedule a feedback for new CLR */
   /************************************/
   if(group_info[group_ind].other_senders[index].is_clr_flag==1)
   {  if(old_is_clr_flag==0) /* this means the receiver becomes a new CLR */
      {  /*****set a timer for 1 rtt *****/
         new_event=(struct time_event_t *) malloc(sizeof(struct time_event_t));
         memset(new_event, 0, sizeof(struct time_event_t));
         if(new_event==NULL)
         {  printf(" error in malloc()\n");
            exit(-1);
         }

         new_event->active_time=get_timestamp() + 
                        group_info[group_ind].other_senders[index].feedback_list.rtt;
         new_event->index=index;
         new_event->fb_nr=bundle_header.fb_nr;

#ifdef UNIX
         pthread_mutex_unlock(&(group_info[group_ind].other_senders_mutex));
         pthread_mutex_lock(&(group_info[group_ind].time_event_mutex));
#else
         ReleaseMutex(group_info[group_ind].other_senders_mutex);
         WaitForSingleObject(group_info[group_ind].time_event_mutex, INFINITE);
#endif
          /* if the time_event link list is empty  */
         if(group_info[group_ind].time_event_tail==NULL)
         {  group_info[group_ind].time_event_head=new_event;
            group_info[group_ind].time_event_tail=new_event;
#ifdef UNIX
            pthread_cond_signal(&(group_info[group_ind].time_event));
#else
            SetEvent(group_info[group_ind].time_event);
#endif
         }
         else   /* if there is already event in the link list */
         {  if(new_event->active_time < 
                        group_info[group_ind].time_event_head->active_time)
            {  if((group_info[group_ind].time_event_head->active_time - 
                                                       new_event->active_time) <32768)
               {  new_event->next=group_info[group_ind].time_event_head;
                  group_info[group_ind].time_event_head=new_event;
#ifdef UNIX
                  pthread_cond_signal(&(group_info[group_ind].time_event));
#else
                  SetEvent(group_info[group_ind].time_event);
#endif
               }
               else
               {  time_search=group_info[group_ind].time_event_head;
                  while(time_search->next != NULL)
                  {  if(time_search->next->active_time > new_event->active_time)
                     {  if((time_search->next->active_time - new_event->active_time)
                                                                           <32768)
                           break;
                     }
                     else
                     {  if((new_event->active_time - time_search->next->active_time)
                                                                           >32768)
                           break;
                     }
                     time_search=time_search->next;
                  }
                  /* at this point, time_search point to either the tail of the
                   * link list or the appropriate position
                   */

                  if(time_search->next ==NULL)  /*search to the tail of the link list*/
                  {  time_search->next=new_event;
                     group_info[group_ind].time_event_tail=new_event;
                  }
                  else
                  {  new_event->next=time_search->next;
                     time_search->next=new_event;
                  }
               }
            }
            else
            {  if((new_event->active_time - 
                           group_info[group_ind].time_event_head->active_time) >32768)
               {  new_event->next=group_info[group_ind].time_event_head;
                  group_info[group_ind].time_event_head=new_event;
#ifdef UNIX
                  pthread_cond_signal(&(group_info[group_ind].time_event));
#else
                  SetEvent(group_info[group_ind].time_event);
#endif
               }
               else
               {  time_search=group_info[group_ind].time_event_head;
                  while(time_search->next != NULL)
                  {  if(time_search->next->active_time > new_event->active_time)
                        break;
                  }
                  /* at this point, time_search point to either the tail of the link
                   * list or the appropriate position
                   */

                  if(time_search->next ==NULL)  /*search to the tail of the link list */
                  {  time_search->next=new_event;
                     group_info[group_ind].time_event_tail=new_event;
                  }
                  else
                  {  new_event->next=time_search->next;
                     time_search->next=new_event;
                  }
               }
            }
         }
#ifdef UNIX
         pthread_mutex_unlock(&(group_info[group_ind].time_event_mutex));
#else
         ReleaseMutex(group_info[group_ind].time_event_mutex);
#endif

      } /* note when the thread deals with the time event, if a event is generated
         * by CLR, then make sure there is only one CLR's event in the link list.
         */
      else
      {
#ifdef UNIX
         pthread_mutex_unlock(&(group_info[group_ind].other_senders_mutex));
#else
         ReleaseMutex(group_info[group_ind].other_senders_mutex);
#endif
      }

   }
   else
   {  /****************************************************/
      /* not CLR, but a new fb_nr, also schedule feedback */
      /****************************************************/
      if(new_fbnr==1)
      {  /***** set a timer as tfmcc draft 4.5 ***/
         new_event=(struct time_event_t *) malloc(sizeof(struct time_event_t));
         memset(new_event, 0, sizeof(struct time_event_t));
         if(new_event==NULL)
         {  printf(" error in malloc()\n");
            exit(-1);
         }
         double temp_delay=1+log10((double)rand()/RAND_MAX)/4;
         if(temp_delay <0 )
            temp_delay=0;
         delay=(unsigned short)
            (4*group_info[group_ind].other_senders[index].feedback_list.rtt*temp_delay);
         /*printf("##########  delay=%d\n", delay);*/
         new_event->active_time=get_timestamp() + delay;
         new_event->index=index;
         new_event->fb_nr=bundle_header.fb_nr;

         
#ifdef UNIX
         pthread_mutex_unlock(&(group_info[group_ind].other_senders_mutex));
         pthread_mutex_lock(&(group_info[group_ind].time_event_mutex));
#else
         ReleaseMutex(group_info[group_ind].other_senders_mutex);
         WaitForSingleObject(group_info[group_ind].time_event_mutex, INFINITE);
#endif
         if(group_info[group_ind].time_event_tail==NULL)
         {  /* if the time_event link list is empty */
            group_info[group_ind].time_event_head=new_event;
            group_info[group_ind].time_event_tail=new_event;
#ifdef UNIX
            pthread_cond_signal(&(group_info[group_ind].time_event));
#else
            SetEvent(group_info[group_ind].time_event);
#endif
         }
         else
         {  /* if there is already event in the link list */
            if(new_event->active_time < 
                              group_info[group_ind].time_event_head->active_time)
            {  if(( group_info[group_ind].time_event_head->active_time - 
                                                   new_event->active_time) <32768)
               {  new_event->next=group_info[group_ind].time_event_head;
                  group_info[group_ind].time_event_head=new_event;
                  group_info[group_ind].cancel_flag=1;
#ifdef UNIX
                  /* notify the timer thread to start waitting.*/
                  pthread_cond_signal(&(group_info[group_ind].time_event));
#else
                  /* notify the timer thread to start waitting.*/
                  SetEvent(group_info[group_ind].time_event); 
#endif
               }
               else
               {  time_search=group_info[group_ind].time_event_head;
                  while(time_search->next != NULL)
                  {  if(time_search->next->active_time > new_event->active_time)
                     {  if((time_search->next->active_time - new_event->active_time)
                                                                  <32768)
                        break;
                     }
                     else
                     {  if((new_event->active_time - time_search->next->active_time)
                                                                  >32768)
                        break;
                     }
                     time_search=time_search->next;
                  }
                  /* at this point, time_search pointed to either the tail of the 
                   * link list or the appropriate position
                   */

                  if(time_search->next ==NULL)  /* search to the tail of the link list*/
                  {  time_search->next=new_event;
                     group_info[group_ind].time_event_tail=new_event;
                  }
                  else
                  {  new_event->next=time_search->next;
                     time_search->next=new_event;
                  }
               }

            }
            else
            {  if(( new_event->active_time - 
                        group_info[group_ind].time_event_head->active_time) >32768)
               {  new_event->next=group_info[group_ind].time_event_head;
                  group_info[group_ind].time_event_head=new_event;
#ifdef UNIX
                  pthread_cond_signal(&(group_info[group_ind].time_event));
#else
                  SetEvent(group_info[group_ind].time_event);
#endif
               }
               else
               {  time_search=group_info[group_ind].time_event_head;
                  while(time_search->next != NULL)
                  {  if(time_search->next->active_time > new_event->active_time)
                     {  if((time_search->next->active_time - new_event->active_time)
                                                                        <32768)
                        break;
                     }
                     else
                     {  if((new_event->active_time - time_search->next->active_time)
                                                                        >32768)
                        break;
                     }
                     time_search=time_search->next;
                  }
                  /* at this point, time_search point to either the tail of the 
                   * link list or the appropriate position
                   */

                  if(time_search->next ==NULL) /* search to the tail of the link list */
                  {  time_search->next=new_event;
                     group_info[group_ind].time_event_tail=new_event;
                  }
                  else
                  {  new_event->next=time_search->next;
                     time_search->next=new_event;
                  }
               }

            }
         }
#ifdef UNIX
         pthread_mutex_unlock(&(group_info[group_ind].time_event_mutex));
#else
         ReleaseMutex(group_info[group_ind].time_event_mutex);
#endif
      }
      else
      {
#ifdef UNIX
         pthread_mutex_unlock(&(group_info[group_ind].other_senders_mutex));
#else
         ReleaseMutex(group_info[group_ind].other_senders_mutex);
#endif
      }

   }

   return 0;

}

/**************************************************************************************/
/* This thread controls when to send out feedback to the other senders in the group.  */
/* All the scheduled feedbacks are stored in a link list:group_info[].time_event_head */
/* upon_receive_bundle() will add in new records into this link list.                 */
/* the scheduled feedback may became obsolete before they are going to be transmitted.*/
/*    1. retrieve a record from link list time_event_head                             */
/*    2. check if the feedback has been cancelled                                     */
/*    3. calculate the receiver desired sending rate                                  */
/*    4. fill in the feedback header                                                  */
/*    5. send out the feedback                                                        */
/*    6. if it is a CLR, then schedule another feedback for the next RTT              */
/**************************************************************************************/
#ifdef UNIX
void * timer_thread(void * mp)
#else
void timer_thread(void * mp)
#endif
{  int i;
   unsigned short int delay;
   int index;
   struct feedback_header_t feedback_header;
   struct feedback_hdr_t feedback_hdr;
   struct time_event_t * new_event;
   double mean_loss_interval=0;
   double mean_loss_interval1=0;
   double mean_loss_interval2=0;
   double desired_rate;
   double p, q; /* the loss event rate, equals to 1/mean_loss_interval */
   struct time_event_t * temp;
   unsigned int sock_srmp;
   sock_srmp=((struct sock_ind_t *)mp)->socket_id;
   int group_ind=((struct sock_ind_t *)mp)->group_index;
   unsigned long int mca;
   int no_loss_num;
   unsigned short int no_loss_period;
   struct time_event_t * time_search;
#ifdef UNIX
   struct timespec delay_time;
#endif


    char tmp = 0;

   /*fprintf(stderr,"Before setting socket option for disable loopback\n");*/

   if (setsockopt(sock_srmp, IPPROTO_IP, IP_MULTICAST_LOOP, (char *)&tmp,
                     sizeof(tmp)) < 0)
   {  fprintf(stderr, "Error: multicast loop back failed\n");
         exit(1);
   }


   mca=tfmcc_mca[group_ind];


#ifdef UNIX
   pthread_mutex_lock(&(group_info[group_ind].time_event_mutex));
#else
   WaitForSingleObject(group_info[group_ind].time_event_mutex, INFINITE);
#endif

   while(1)
   {  /*printf("-");
      fflush(stdout);*/
#ifndef UNIX
      ResetEvent(group_info[group_ind].time_event);
#endif
      if(group_info[group_ind].time_event_head==NULL)
      {  /* there is no entry in the event link list */
         /*printf("point 1000 in timer thread\n");*/
       /****************************************************/
       /* waiting for the time_event_head became non empty */
       /****************************************************/
#ifdef UNIX
         pthread_cond_wait(&(group_info[group_ind].time_event),
                                    &(group_info[group_ind].time_event_mutex));
#else
         ReleaseMutex(group_info[group_ind].time_event_mutex);
         /* wait for the event to be set */
         WaitForSingleObject(group_info[group_ind].time_event, INFINITE);
         WaitForSingleObject(group_info[group_ind].time_event_mutex, INFINITE);
#endif
         /*printf("point 2000 in timer thread\n"); */
      }
      else
      {   /*printf("point 3000 in timer thread\n");*/
         /******************************************/
         /* retrieve a record from time_event_head */
         /******************************************/
         index=group_info[group_ind].time_event_head->index;
         delay=group_info[group_ind].time_event_head->active_time-get_timestamp();

/*         printf("active time=%d, current time=%d, delay=%d\n", 
          group_info[group_ind].time_event_head->active_time,  get_timestamp(), delay);
         if(delay>32768) */
         if(delay>500)
         {  delay=0;
            printf("active time= %d, current_time = %d, delay = %d\n",
                          group_info[group_ind].time_event_head->active_time, 
                          get_timestamp(), delay);
            /*printf("multi thread performance problem\n");*/
         }
         /*printf("point 40000 inside timer_thread\n");*/
#ifdef UNIX
         pthread_mutex_unlock(&(group_info[group_ind].time_event_mutex));
         pthread_mutex_lock(&(group_info[group_ind].other_senders_mutex));
         pthread_mutex_lock(&(group_info[group_ind].time_event_mutex));
#else
         /* note, dangeroous for deadlock */
         ReleaseMutex(group_info[group_ind].time_event_mutex);
         WaitForSingleObject(group_info[group_ind].other_senders_mutex, INFINITE);
         WaitForSingleObject(group_info[group_ind].time_event_mutex, INFINITE);
#endif

         if(group_info[group_ind].time_event_head->fb_nr==
                        group_info[group_ind].other_senders[index].sender_list.fb_nr)
         {
#ifdef UNIX
            pthread_mutex_unlock(&(group_info[group_ind].time_event_mutex));
            delay_time.tv_sec=delay/1000;
            delay_time.tv_nsec=(delay %1000)*1000000;
            pthread_cond_reltimedwait_np(&(group_info[group_ind].time_event),
                           &(group_info[group_ind].other_senders_mutex), &delay_time);
            pthread_mutex_lock(&(group_info[group_ind].time_event_mutex));
#else
            ReleaseMutex(group_info[group_ind].time_event_mutex);
            ReleaseMutex(group_info[group_ind].other_senders_mutex);
            WaitForSingleObject(group_info[group_ind].time_event, delay);
            WaitForSingleObject(group_info[group_ind].other_senders_mutex, INFINITE);
            WaitForSingleObject(group_info[group_ind].time_event_mutex, INFINITE);
#endif
         }

         if(group_info[group_ind].cancel_flag==1) 
         {  /* the current timer was cancelled
             * this may be cause eigher by a new short event inserted before
             * or the receiver_leave_group();
             */
            /********************************************/
            /* current feedback was cancelled, next one  */
            /********************************************/
            group_info[group_ind].cancel_flag=0;
#ifdef UNIX
            pthread_mutex_unlock(&(group_info[group_ind].other_senders_mutex));
#else
            ReleaseMutex(group_info[group_ind].other_senders_mutex);
#endif
            continue;
         }


         /***********************************************************/
         /* calculate the receiver desired rate, as tfmcc draft 5.4 */
         /***********************************************************/
         if(group_info[group_ind].other_senders[index].have_loss_flag==0) 
         {   /* there is no packet loss, double the current receive rate */
            desired_rate=group_info[group_ind].other_senders[index].x_recv*2;
         }
         else if(group_info[group_ind].time_event_head->fb_nr==
                        group_info[group_ind].other_senders[index].sender_list.fb_nr)
         {  mean_loss_interval=0;
            for(i=0; i<8; i=i+1)
            {  mean_loss_interval=mean_loss_interval+weight[i]*
                        group_info[group_ind].other_senders[index].loss_interval[i+2];
            }
            mean_loss_interval=mean_loss_interval/6;

            if(mean_loss_interval == 0)
            {  /* Initializing the loss history after the first loss event, 
                * as tfmcc draft 5.6
                */
               group_info[group_ind].other_senders[index].loss_interval[9]=
                    (int)(group_info[group_ind].other_senders[index].x_recv
                     *group_info[group_ind].other_senders[index].feedback_list.rtt
                     /1000/sqrt(2.0/3.0)
                     /group_info[group_ind].other_senders[index].mean_packet_size);
               group_info[group_ind].other_senders[index].loss_interval[9]=
                (int)pow(group_info[group_ind].other_senders[index].loss_interval[9],2);
               mean_loss_interval=
                        group_info[group_ind].other_senders[index].loss_interval[9];
               /*printf("mean_loss_interval=%f\n", mean_loss_interval);*/
            }
            else
            {  /* the second case is: after the most recent packet loss, there is no
                * futhur packet loss. and this no loss last a significent long period of
                * time. During this period, the X_r should be increased if possible.
                */
               if(group_info[group_ind].other_senders[index].packet_history[19].sn 
                                    >group_info[group_ind].other_senders[index].s_old)
                  no_loss_num=
                     group_info[group_ind].other_senders[index].packet_history[19].sn
                     - group_info[group_ind].other_senders[index].s_old;
               else   /* sn wrapped */
                  no_loss_num=
                     group_info[group_ind].other_senders[index].packet_history[19].sn
                     +65536- group_info[group_ind].other_senders[index].s_old;

               mean_loss_interval1=0;
               for(i=0; i<7; i=i+1)
               {  mean_loss_interval1=mean_loss_interval1+weight[i]*
                        group_info[group_ind].other_senders[index].loss_interval[i+3];
               }
               mean_loss_interval1=mean_loss_interval1+weight[7]*no_loss_num;
               mean_loss_interval1=mean_loss_interval1 / 6;

               /* the third case: previous feedback leads to the sending rate decreased
                * to a very small value. since then, no packet loss, and few new packet
                * arrived. During this period, the sending rate should be increased.
                * Note: This is not specified in the TFMCC draft.
                */
               no_loss_period=get_timestamp() - 
                              group_info[group_ind].other_senders[index].t_old;
               no_loss_num=no_loss_period /
                        group_info[group_ind].other_senders[index].feedback_list.rtt;

               mean_loss_interval2=0;
               for(i=0; i<7; i=i+1)
               {  mean_loss_interval2=mean_loss_interval2+weight[i]*
                        group_info[group_ind].other_senders[index].loss_interval[i+3];
               }
               mean_loss_interval2=mean_loss_interval2+weight[7]*no_loss_num;
               mean_loss_interval2=mean_loss_interval2 / 6;

            }

            /* At this point, need to compare mean_loss_interval 1,2,and 3, 
             *  get the maximum value 
             */
            if(mean_loss_interval < mean_loss_interval1)
               mean_loss_interval=mean_loss_interval1;

            if(mean_loss_interval < mean_loss_interval2)
               mean_loss_interval=mean_loss_interval2;


            /* calculate the desired rate as TCP throughput equation */
            p=1/mean_loss_interval;
            q=(sqrt(p*2/3)+p*12*sqrt(p*3/8)*(1+32*p*p));
            desired_rate=(double)8*
                  group_info[group_ind].other_senders[index].mean_packet_size
                  /group_info[group_ind].other_senders[index].feedback_list.rtt*1000;
            /*printf("desired_rate=%f\n", desired_rate);*/
            desired_rate=desired_rate/q;
            /*   printf("p=%f, q=%f,  desired_rate=%f\n",p,q,  desired_rate);*/

         }

         /*   for(i=0; i<8; i=i+1)
         {  printf("%d  ", 
                  group_info[group_ind].other_senders[index].loss_interval[i+2]);
         }*/
         /*printf("\nmean loss interval= %f \n", mean_loss_interval);*/


         /**********************************************/
         /* send out feedback packet( maybe supressed) */
         /**********************************************/
         if(group_info[group_ind].other_senders[index].is_clr_flag==1)
         {  /* if it is clr, send feedback, and schedule another feedback after RTT*/
            /* the current receiver is the CLR, send out the feedback
             * and schedule another feedback that will be actived 1 RTT later
             */
            unsigned short int temp123;
            temp123=(int)
               group_info[group_ind].other_senders[index].sender_list.sender_timestamp 
               +(get_timestamp()- 
                 group_info[group_ind].other_senders[index].sender_list.received_time);
            feedback_header.sender_timestamp=temp123;
            strcpy(feedback_header.sender_id, 
                  group_info[group_ind].other_senders[index].sender_list.sender_id);
            strcpy(feedback_header.receiver_id, 
                           group_info[group_ind].other_senders[index].receiver_id);
            feedback_header.fb_nr=
                        group_info[group_ind].other_senders[index].sender_list.fb_nr;
            feedback_header.receiver_timestamp=get_timestamp();
            feedback_header.have_rtt_flag=
                 group_info[group_ind].other_senders[index].feedback_list.have_rtt_flag;
            feedback_header.have_loss_flag=
                        group_info[group_ind].other_senders[index].have_loss_flag;
            feedback_header.x_r=desired_rate;
            feedback_header.receiver_leave_flag=
                                       group_info[group_ind].current_receiver_leave;

            assemble_feedback_header(feedback_header, &feedback_hdr);
            if(group_info[group_ind].current_receiver_leave==1)
               group_info[group_ind].other_senders[index].leave_rtt_count=
                           group_info[group_ind].other_senders[index].leave_rtt_count+1;

            /* remove the record in group_info[group_ind].time_event_head */
            temp=group_info[group_ind].time_event_head;
            group_info[group_ind].time_event_head=
                                    group_info[group_ind].time_event_head->next;
            if(group_info[group_ind].time_event_head==NULL)
               group_info[group_ind].time_event_tail=NULL;
            free(temp);

            if(group_info[group_ind].other_senders[index].leave_rtt_count<=4)
            {  /* the current CLR is not leaving
                * or is leaving, but within 4 RTT
                */
               /*show_feedback_header(feedback_header);*/
               send_feedback(sock_srmp, mca, feedback_hdr);
               /*printf("   *** sent out feedback because of being a CLR at time %d\n",
                                  get_timestamp());
               printf(".");
               show_feedback_header(feedback_header);
               printf("\n");
               */
               /*****set a timer for 1 rtt *****/
               new_event=(struct time_event_t *) malloc(sizeof(struct time_event_t));
               memset(new_event, 0, sizeof(struct time_event_t));
               if(new_event==NULL)
               {  printf(" error in malloc()\n");
                  exit(-1);
               }

               new_event->active_time=get_timestamp() + 
                           group_info[group_ind].other_senders[index].feedback_list.rtt;
               new_event->index=index;
               new_event->fb_nr=
                           group_info[group_ind].other_senders[index].sender_list.fb_nr;
               /*printf("next active_time=%d\n", new_event->active_time);*/
               /*note: the original one is new_event->fb_nr=bundle_header.fb_nr;*/

               if(group_info[group_ind].time_event_tail==NULL) 
               {  /* if the time_event link list is empty */
                  group_info[group_ind].time_event_head=new_event;
                  group_info[group_ind].time_event_tail=new_event;

               }
               else
               {  /* if there is already event in the link list  */
                  if(new_event->active_time < 
                           group_info[group_ind].time_event_head->active_time)
                  {  if(( group_info[group_ind].time_event_head->active_time - 
                                                      new_event->active_time) <32768)
                     {  new_event->next=group_info[group_ind].time_event_head;
                        group_info[group_ind].time_event_head=new_event;
                     }
                     else
                     {  time_search=group_info[group_ind].time_event_head;
                        while(time_search->next != NULL)
                        {  if(time_search->next->active_time > new_event->active_time)
                           {  if((time_search->next->active_time - 
                                       new_event->active_time) <32768)
                                 break;
                           }
                           else
                           {  if((new_event->active_time - 
                                       time_search->next->active_time) >32768)
                                 break;
                           }

                           time_search=time_search->next;
                        }
                        /* at this point, time_search point to either the tail of 
                         *the link list or the appropriate position
                         */

                        if(time_search->next ==NULL)
                        {  /* search to the tail of the link list */
                           time_search->next=new_event;
                           group_info[group_ind].time_event_tail=new_event;
                        }
                        else
                        {  new_event->next=time_search->next;
                           time_search->next=new_event;
                        }
                     }

                  }
                  else
                  {  if(( new_event->active_time - 
                           group_info[group_ind].time_event_head->active_time) >32768)
                     {  new_event->next=group_info[group_ind].time_event_head;
                        group_info[group_ind].time_event_head=new_event;
                     }
                     else
                     {  time_search=group_info[group_ind].time_event_head;
                        while(time_search->next != NULL)
                        {  if(time_search->next->active_time > new_event->active_time)
                           {  if((time_search->next->active_time - 
                                                   new_event->active_time) <32768)
                              break;
                           }
                           else
                           {  if((new_event->active_time - 
                                             time_search->next->active_time) >32768)
                              break;
                           }
                           time_search=time_search->next;
                        }
                        /* at this point, time_search point to either the tail of the
                         * link list or the appropriate position
                         */

                        if(time_search->next ==NULL)
                        {  /* search to the tail of the link list */
                           time_search->next=new_event;
                           group_info[group_ind].time_event_tail=new_event;
                        }
                        else
                        {  new_event->next=time_search->next;
                           time_search->next=new_event;
                        }
                     }



                  }
               } /* end of set RTT timer */
            }
            else
            {  /* current receiver is CLR, is leaving
                * and 4 RTT expired
                * can leave safely
                */
               /* clean up group_info[group_ind].other_senders[index] */
               memset(&group_info[group_ind].other_senders[index], 0, 
                                             sizeof(struct other_sender_t));
               group_info[group_ind].other_senders[index].s_old=-1;
               group_info[group_ind].other_senders[index].have_loss_flag=0;
               group_info[group_ind].other_senders[index].minimum_x_supp=1e20;
               strcpy(group_info[group_ind].other_senders[index].receiver_id, 
                                                                     ip_address);
               group_info[group_ind].other_senders[index].sender_list.fb_nr=-1;
               group_info[group_ind].other_senders[index].mean_packet_size=-1;
               group_info[group_ind].other_senders[index].number_of_loss=0;
               for(i=0; i<5; i=i+1)
               {  group_info[group_ind].other_senders[index].packet_history[i].sn=-1;
                  group_info[group_ind].other_senders[index].
                                                packet_history[i].received_time=65535;
               }

               /* check to see if the receiver can leave safely */
               for(i=0; i<20; i=i+1)
               {  if(group_info[group_ind].other_senders[i].sender_list.fb_nr>=0)
                     break;
               }
               if(i>=20)
               {  printf("\n\n   *** the receiver can leave safely ***\a\a\a\n\n\n");
               }

            }

         }
         /* the current receiver is not the CLR */
         else if(group_info[group_ind].time_event_head->fb_nr==
                  group_info[group_ind].other_senders[index].sender_list.fb_nr)
         {  /*printf("rtt=%d , r_max=%d\n",
                group_info[group_ind].other_senders[index].feedback_list.rtt,
                group_info[group_ind].other_senders[index].sender_list.r_max);
            printf("desired rate = %f, x_supp=%f\n", desired_rate, 
                group_info[group_ind].other_senders[index].sender_list.x_supp);
            */
            if(desired_rate <
               group_info[group_ind].other_senders[index].sender_list.x_supp ||
               group_info[group_ind].other_senders[index].feedback_list.rtt>
               group_info[group_ind].other_senders[index].sender_list.r_max)
            /* the desired rate of the receiver is less than the rate specified 
             * in the bundle header or the measured rtt is large than the r_max
             * specified in the bundle header
             * can send out feedback
             */
            if(desired_rate < group_info[group_ind].other_senders[index].minimum_x_supp)
            {  int temp;
               /*printf("minimum_x_supp=%f\n", 
                      group_info[group_ind].other_senders[index].minimum_x_supp);*/
               unsigned short int current=get_timestamp();
               temp=(int)group_info[group_ind].other_senders[index].sender_list.
                                                            sender_timestamp + current;
               temp=temp-
                  group_info[group_ind].other_senders[index].sender_list.received_time;
               feedback_header.sender_timestamp=temp;
               strcpy(feedback_header.sender_id, 
                     group_info[group_ind].other_senders[index].sender_list.sender_id);
               strcpy(feedback_header.receiver_id,
                               group_info[group_ind].other_senders[index].receiver_id);
               feedback_header.fb_nr=
                          group_info[group_ind].other_senders[index].sender_list.fb_nr;
               feedback_header.receiver_timestamp=get_timestamp();
               feedback_header.have_rtt_flag=
                 group_info[group_ind].other_senders[index].feedback_list.have_rtt_flag;
               feedback_header.have_loss_flag=
                           group_info[group_ind].other_senders[index].have_loss_flag;
               feedback_header.x_r=desired_rate;
               feedback_header.receiver_leave_flag=
                                          group_info[group_ind].current_receiver_leave;
               /* printf("feedback sent out time=%d\n", current); */
               assemble_feedback_header(feedback_header, &feedback_hdr);
               /*show_feedback_header(feedback_header);*/
               send_feedback(sock_srmp, mca, feedback_hdr);
               /*printf("   *** sent out feedback\n");
                show_feedback_header(feedback_header);
                printf("\n");
                */
            }

            /* remove the record in group_info[group_ind].time_event_head */
            temp=group_info[group_ind].time_event_head;
            group_info[group_ind].time_event_head=
                                          group_info[group_ind].time_event_head->next;
            if(group_info[group_ind].time_event_head==NULL)
               group_info[group_ind].time_event_tail=NULL;
            free(temp);
         }
         /*note: this feature need to be done in SRMP side*/
         /*feedback_header.receiver_leave_flag=0;*/

         else
         {  /* remove the record in group_info[group_ind].time_event_head */
            temp=group_info[group_ind].time_event_head;
            group_info[group_ind].time_event_head=
                                    group_info[group_ind].time_event_head->next;
            if(group_info[group_ind].time_event_head==NULL)
               group_info[group_ind].time_event_tail=NULL;
            free(temp);
         }
#ifdef UNIX
         pthread_mutex_unlock(&(group_info[group_ind].other_senders_mutex));
#else
         ReleaseMutex(group_info[group_ind].other_senders_mutex);
#endif
      }

   } /* end of while; */

} /* end of timer_thread(); */



/************************************************************/
/* the function to print the content inside a bundle header */
/************************************************************/
void show_bundle_header(struct bundle_header_t * bundle_header)
{  printf("   bundle_header->sender_id= %s\n", bundle_header->sender_id);
   printf("   bundle_header->sn= %d\n", bundle_header->sn);
   printf("   bundle_header->fb_nr= %d\n", bundle_header->fb_nr);
   printf("   bundle_header->r_max= %d\n", bundle_header->r_max);
   printf("   bundle_header->received_time= %d\n", bundle_header->received_time);
   printf("   bundle_header->receiver_id= %s\n", bundle_header->receiver_id);
   printf("   bundle_header->receiver_timestamp= %d\n",
                                                   bundle_header->receiver_timestamp);
   printf("   bundle_header->is_clr_flag= %d\n", bundle_header->is_clr_flag);
   printf("   bundle_header->sender_timestamp= %d\n", bundle_header->sender_timestamp);
   printf("   bundle_header->x_supp= %12.4g\n", bundle_header->x_supp);
   printf("   bundle_header->dsn_count= %d\n", bundle_header->dsn_count);
   printf("   bundle_header->len= %d\n", bundle_header->len);
   return;
}


/*******************************************************/
/* the function to print the content inside a feedback */
/*******************************************************/
void show_feedback_header(struct feedback_header_t  feedback_header)
{  printf("   feedback_header->receiver_id= %s\n", feedback_header.receiver_id);
   printf("   feedback_header->sender_id= %s\n", feedback_header.sender_id);
   printf("   feedback_header->fb_nr= %d\n", feedback_header.fb_nr);
   printf("   feedback_header->x_r= %f\n", feedback_header.x_r);
   printf("   feedback_header->sender_timestamp= %d\n", 
                                                   feedback_header.sender_timestamp);
   printf("   feedback_header->received_time= %d\n", feedback_header.received_time);
   printf("   feedback_header->receiver_timestamp= %d\n", 
                                                feedback_header.receiver_timestamp);
   printf("   feedback_header->have_rtt_flag= %d\n", feedback_header.have_rtt_flag);
   printf("   feedback_header->have_loss_flag= %d\n", feedback_header.have_loss_flag);
   printf("   feedback_header->receiver_leave_flag= %d\n", 
                                                feedback_header.receiver_leave_flag);
   return;
}


/********************************************************************************/
/* SRMP will call this function to indicate the current receiver is going to    */
/*  leave the multicast group                                                   */
/* This may happen when no clients are subscribing nor registering the group    */
/* After call this, SRMP should not be terminated immediately, because if the   */
/*  receiver is a CLR, it should not leave the group until other receiver       */
/*  became the new CLR, or 4 RTT period expired                                 */
/* This function will not clear up everything in group_info[], only stop the    */
/*  TFMCC from sending furthur feedback                                         */
/*   1. remove all the feedback that current receiver act as a CLR              */
/*   2. clear up group_info[].other_senders[]                                   */
/*   3. wait all the remote sender's permit, that is, if the the current        */
/*      receiver is a CLR for a remote sender, then the receiver will wait for  */
/*      another bundle from the sender. If the CLR changed to another one, the  */
/*      the current receiver can safely leave the sender. Or else, the current  */
/*      receiver eait for 4 RTTs, and then leave                                */
/********************************************************************************/
int receiver_leave_group(unsigned long int mca)
{  int i,j;
   struct time_event_t * temp_time_event;
   struct time_event_t * temp2_time_event;
   int index;
   int group_ind;
   group_ind=search_tfmcc_mca(mca, 0);

   temp_time_event=NULL;
   temp2_time_event=NULL;

   /* note here: waitting for two mutex may leads to deadlock!!! */
#ifdef UNIX
   pthread_mutex_lock(&(group_info[group_ind].other_senders_mutex));
   pthread_mutex_lock(&(group_info[group_ind].time_event_mutex));
#else
   WaitForSingleObject(group_info[group_ind].other_senders_mutex, INFINITE);
   WaitForSingleObject(group_info[group_ind].time_event_mutex, INFINITE);
#endif

   group_info[group_ind].current_receiver_leave=1;

   /*************************************************************************/
   /* first remove all the record in group_info[group_ind].time_event_head, */
   /* that the receiver is a CLR                                            */
   /*************************************************************************/

   if(group_info[group_ind].time_event_head!=NULL)
   {  index=group_info[group_ind].time_event_head->index;
      if (group_info[group_ind].other_senders[index].is_clr_flag==0)
         /* the first time event */
         group_info[group_ind].cancel_flag=1;

      /* move the link list head to the first record does not belong to a CLR */
      temp_time_event=group_info[group_ind].time_event_head;
      while( (temp_time_event!=NULL) && 
                     (group_info[group_ind].other_senders[index].is_clr_flag==0) )
      {  group_info[group_ind].time_event_head=temp_time_event->next;
         free(temp_time_event);
         temp_time_event=group_info[group_ind].time_event_head;
         if(temp_time_event!=NULL)
            index=temp_time_event->index;
      }
      if(temp_time_event==NULL)
         group_info[group_ind].time_event_tail=NULL;

      /* then remove all the remainning records that belong to a CLR */
      if(temp_time_event!=NULL)
         temp2_time_event=temp_time_event->next;
      if(temp2_time_event!=NULL)
         index=temp2_time_event->index;
      while(1)
      {  if((temp2_time_event!=NULL) && 
                     (group_info[group_ind].other_senders[index].is_clr_flag==0) )
         {  temp2_time_event=temp2_time_event->next;
            free(temp_time_event->next);
            temp_time_event->next=temp2_time_event;
         }
         else if(temp2_time_event!=NULL)
         {  temp2_time_event=temp2_time_event->next;
            temp_time_event=temp_time_event->next;
         }
         else
         {  break;
         }

         if(temp2_time_event!=NULL)
            index=temp2_time_event->index;
      }
      group_info[group_ind].time_event_tail=temp_time_event;
   }
#ifdef UNIX
   pthread_mutex_unlock(&(group_info[group_ind].time_event_mutex));
#else
   ReleaseMutex(group_info[group_ind].time_event_mutex);
#endif
   /* printf("point 7000 in receiver leave group\n"); */
   /************************************************************************/
   /* second clear up group_info[].other_senders[] for the indexed sender  */
   /************************************************************************/
   for(j=0; j<20; j=j+1)
   {  if(group_info[group_ind].other_senders[j].is_clr_flag==0)
      {  memset(&group_info[group_ind].other_senders[j], 0, 
                                                      sizeof(struct other_sender_t));
         group_info[group_ind].other_senders[j].s_old=-1;
         group_info[group_ind].other_senders[j].have_loss_flag=0;
         group_info[group_ind].other_senders[j].minimum_x_supp=1e20;
         strcpy(group_info[group_ind].other_senders[j].receiver_id, ip_address);
         group_info[group_ind].other_senders[j].sender_list.fb_nr=-1;
         group_info[group_ind].other_senders[j].number_of_loss=0;

         for(i=0; i<20; i=i+1)
         {  group_info[group_ind].other_senders[j].packet_history[i].sn=-1;
            group_info[group_ind].other_senders[j].packet_history[i].received_time
                                                                                 =65535;
         }
      }
   }

   /************************************************/
   /* third, check to see if no sender to wait.    */
   /* If yes, send signal to inform safely leave   */
   /************************************************/
   for(i=0; i<20; i=i+1)
   {  if(group_info[group_ind].other_senders[i].sender_list.fb_nr>=0)
         break;
   }
   if(i>=20)
   {  if(group_info[group_ind].time_event_head==NULL)
      {   printf("\n\n   *** the receiver can leave safely ***\a\a\a\n\n\n");
      }
   }
#ifdef UNIX
   pthread_mutex_unlock(&(group_info[group_ind].other_senders_mutex));
#else
   ReleaseMutex(group_info[group_ind].other_senders_mutex);
#endif
   /* printf("point 9000 in receiver leave group\n"); */
   return 0;
}


/**************************************************************************************/
/*   This function is to resume the receiver in multicast group                       */
/*   After SRMP called receiver_leave(), if SRMP need to rejoin the multicast group,  */
/*     this function call will resume TFMCC function for the specified group          */
/**************************************************************************************/
int receiver_resume_group(unsigned long int mca)
{  int group_ind;
   group_ind=search_tfmcc_mca(mca, 0);
#ifdef UNIX
   pthread_mutex_lock(&(group_info[group_ind].other_senders_mutex));
#else
   WaitForSingleObject(group_info[group_ind].other_senders_mutex, INFINITE);
#endif
   group_info[group_ind].current_receiver_leave=0;
#ifdef UNIX
   pthread_mutex_unlock(&(group_info[group_ind].other_senders_mutex));
#else
   ReleaseMutex(group_info[group_ind].other_senders_mutex);
#endif
   return 0;
}


/**************************************************************************************/
/* This function should be called by SRMP each time SRMP is going to send out a       */
/*  message, either mode 0 or mode 1.                                                 */
/*                                                                                    */
/*    1. update sending history                                                       */
/*    2. calculate mean sending rate during MEAN_SEND_PERIOD                          */
/*    3. if a mode 0 message                                                          */
/*           1. update TFMCC x_send                                                   */
/*           2. calculate mode 0 drop probability                                     */
/*           3. drop mode 0 packet randomly                                           */
/*                                                                                    */
/*  return value: 0 packet should not be dropped                                      */
/*                1 packet should be dropped                                          */
/**************************************************************************************/
int drop_mode_0(int mode, int len, unsigned long int mca)
{  double rate;
   double random_number;
   double drop_probability;
   unsigned short int current_time;

   int i;
   unsigned short int time_thresh;
   unsigned short int time_duration;

   int mode_0_len;
   int mode_1_len;
   int total_len;
   int permit_len;
   int group_ind;
   /*fflush(stdout);*/
   group_ind=search_tfmcc_mca(mca, 0);

   mode_0_len=0;
   mode_1_len=0;
   time_duration=0;

   /*************************/
   /* update packet history */
   /*************************/

   current_time=get_timestamp();
   time_thresh=current_time - MEAN_SEND_PERIOD;

   for(i=0; i<19; i=i+1)
   {  group_info[group_ind].srmp_send_his[i].len=
                                    group_info[group_ind].srmp_send_his[i+1].len;
      group_info[group_ind].srmp_send_his[i].mode=
                                    group_info[group_ind].srmp_send_his[i+1].mode;
      group_info[group_ind].srmp_send_his[i].sending_time=
                           group_info[group_ind].srmp_send_his[i+1].sending_time;

      if(group_info[group_ind].srmp_send_his[i].sending_time < time_thresh)
      {  group_info[group_ind].srmp_send_his[i].len=0;
         time_duration=MEAN_SEND_PERIOD;
      }
      else if ((group_info[group_ind].srmp_send_his[i].sending_time - time_thresh)
                                                                              >32768)
      {  group_info[group_ind].srmp_send_his[i].len=0;
         time_duration=MEAN_SEND_PERIOD;
      }

      if(group_info[group_ind].srmp_send_his[i].mode == 0)
         mode_0_len = mode_0_len + group_info[group_ind].srmp_send_his[i].len;
      else
         mode_1_len = mode_1_len + group_info[group_ind].srmp_send_his[i].len;
   }
   /* the new message */
   group_info[group_ind].srmp_send_his[19].len = len;
   group_info[group_ind].srmp_send_his[19].mode = mode;
   group_info[group_ind].srmp_send_his[19].sending_time = current_time;
   if(mode == 0)
      mode_0_len = mode_0_len + group_info[group_ind].srmp_send_his[19].len;
   else
   {  /* if it is mode 1 packet, should not be dropped.
       * simply return.
       */
      /*fflush(stdout);*/
      return 0;
   }
   
   /* else, mode 0 */
   /* calculate time duration */
   if(time_duration == 0)
      time_duration = current_time - 
                              group_info[group_ind].srmp_send_his[0].sending_time;
   if(group_info[group_ind].srmp_send_his[0].len==0)
      time_duration=MEAN_SEND_PERIOD;
/* for(i=0; i<20; i=i+1)
   {  printf("%d  ", group_info[group_ind].srmp_send_his[i].len);
   }
   printf("\n");

   for(i=0; i<20; i=i+1)
   {  printf("%d  ", group_info[group_ind].srmp_send_his[i].sending_time);
   }
   printf("\n");

*/

   total_len= mode_0_len + mode_1_len;

   /***************************************/
   /* calculate TFMCC rate and permit_len */
   /***************************************/
#ifdef UNIX
   pthread_mutex_lock(&(group_info[group_ind].mutex));
#else
   WaitForSingleObject(group_info[group_ind].mutex, INFINITE);
#endif

   current_time=get_timestamp();

   /* set the sending rate */
   if(group_info[group_ind].x_send < group_info[group_ind].target_rate)
   {  if(group_info[group_ind].x_send<0)
         group_info[group_ind].x_send=0;
      current_time=get_timestamp();
      if(group_info[group_ind].in_slowstart_flag==1)
      {  rate=group_info[group_ind].x_send+(
                  current_time-group_info[group_ind].start_point)
                  *group_info[group_ind].increase_rate;
      }
      else
      {  if(group_info[group_ind].mean_packet_size<0)
            rate=group_info[group_ind].x_send+
               (current_time-group_info[group_ind].start_point)*8*DEFAULT_MEAN_PKT_SIZE/
                  group_info[group_ind].r_max/group_info[group_ind].r_max*1000;
         else
            rate=group_info[group_ind].x_send+
               (current_time-group_info[group_ind].start_point)*8*
               group_info[group_ind].mean_packet_size/group_info[group_ind].r_max/
               group_info[group_ind].r_max*1000;
      }

      group_info[group_ind].start_point=current_time;

      if(rate>=group_info[group_ind].target_rate)
      {  rate=group_info[group_ind].target_rate;
         group_info[group_ind].increase_rate=0;
         group_info[group_ind].target_rate=-1;
         group_info[group_ind].start_point=65535;
      }

      group_info[group_ind].x_send=rate;
   }
   else
   {  rate=group_info[group_ind].x_send;
   }

#ifdef UNIX
   pthread_mutex_unlock(&(group_info[group_ind].mutex));
#else
   ReleaseMutex(group_info[group_ind].mutex);
#endif
   if(rate==-1)  /* sender initialize */
   {  /*fflush(stdout);*/
      return 0;
   }
   if(rate==-0.5) /* 1 packet per 64 seconds */
   {
#ifdef UNIX
      time_t current_seconds;
      time(&current_seconds);
#else
      SYSTEMTIME current_systime;
      unsigned long int current_seconds;
      GetSystemTime(&current_systime);
      current_seconds=((current_systime.wDay*24+current_systime.wHour)*60+
                      current_systime.wMinute)*60+current_systime.wSecond;
#endif
      if(current_seconds<(group_info[group_ind].last_packet_time+64))
         return 1;  /* packet should be dropped */
      else 
      {  group_info[group_ind].last_packet_time=current_seconds;
         return 0;
      }
   }

   permit_len= (int) (rate * time_duration/8/1000);

   /*printf("x_send=%f, target_rate=%f  ", rate,group_info[group_ind].target_rate );
     printf("permit_len=%d, total_len=%d\n",  permit_len, total_len);
     printf("time_duration=%d\n", time_duration);
    */
   /***************************************************************/
   /* 3. check to see if current mode 0 message should be dropped */
   /***************************************************************/
   /* if permited len is more than total sent
    * current message should not be dropped
    */
   if(permit_len > total_len)
   {   /*fflush(stdout);*/
       return 0;
    }

   /* if permited len is less than mode 1 len sent
    *   current message should be dropped
    */
   if(permit_len < mode_1_len)
    {  /*fflush(stdout);*/
       return 1;
    }
   /*  otherwise the message should be dropped randomly */
   drop_probability=(double)(total_len - permit_len) / total_len;

   random_number=(double)rand()/RAND_MAX;

   if(drop_probability <random_number)
   {  /*fflush(stdout);*/
      return 0;
   }

   else
   {  /*fflush(stdout);*/
      return 1;
   }
}

/**************************************************************************************/
/* given struct bundle_header_t, fill in struct bundle_hdr_t, which is the actual     */
/* bundle header used during transmission                                             */
/**************************************************************************************/
int assemble_bundle_header(struct bundle_header_t bundle_header_para,
                                                   struct bundle_hdr_t * bundle_header)
{  int a, b, m;
#ifndef UNIX
   __u8 * byte_pointer;
   __u8 temp_byte1;
#endif

   bundle_header->version= 0x2;
   bundle_header->type= 0x0;
   bundle_header->fb_nr=bundle_header_para.fb_nr;
   bundle_header->is_clr_flag=bundle_header_para.is_clr_flag;
   bundle_header->bundle_sn=htons(bundle_header_para.sn);
   bundle_header->sender_id.s_addr=inet_addr(bundle_header_para.sender_id);
   bundle_header->receiver_id.s_addr=inet_addr(bundle_header_para.receiver_id);
   bundle_header->sender_timestamp=htons(bundle_header_para.sender_timestamp);
   bundle_header->receiver_timestamp=htons(bundle_header_para.receiver_timestamp);
   bundle_header->r_max=htons(bundle_header_para.r_max);
   bundle_header->dsn_count=bundle_header_para.dsn_count;
   bundle_header->len=htons(bundle_header_para.len);

   if(bundle_header_para.x_supp>2047e31)
   {  bundle_header->x_supp_man=2047;
      bundle_header->x_supp_exp=31;
   }
   else
   {  /* convert x_supp from a double into 16bit */
      m=(int) (log(bundle_header_para.x_supp)/log(10) + 1);
      a=(int) (bundle_header_para.x_supp/pow(10,(m-4)));
      b=m-4;
      if(b<0)
      {  a=(int)bundle_header_para.x_supp;
         b=0;
      }
      else if(a>=2048)
      {  a=a/10;
         b=b+1;
      }
      bundle_header->x_supp_man=a;
      bundle_header->x_supp_exp=b;
   }

#ifndef UNIX
   /* transfrom the bit sequence from windows to UNIX */
   byte_pointer=(__u8 *)bundle_header;
   temp_byte1=*(byte_pointer+16);
   *(byte_pointer+16)=*(byte_pointer+17);
   *(byte_pointer+17)=temp_byte1;
#endif

   return 0;
}

/**************************************************************************************/
/* given struct bundle_hdr_t, fill in struct bundle_header_t, which will be used as   */
/* parameter when TFMCC processing bundle information                                 */
/**************************************************************************************/
int dessemble_bundle_header(struct bundle_header_t * bundle_header_para, 
                                          struct bundle_hdr_t bundle_header)
{
#ifndef UNIX
   __u8 * byte_pointer;
   __u8 temp_byte1;
   /* transform from Unix bit sequency into Windows sequency */
   byte_pointer=(__u8 *)&bundle_header;
   temp_byte1=*(byte_pointer+16);
   *(byte_pointer+16)=*(byte_pointer+17);
   *(byte_pointer+17)=temp_byte1;
#endif

   bundle_header_para->fb_nr=bundle_header.fb_nr;
   bundle_header_para->is_clr_flag=bundle_header.is_clr_flag;
   bundle_header_para->r_max=ntohs(bundle_header.r_max);
   strcpy( bundle_header_para->receiver_id, inet_ntoa(bundle_header.receiver_id) );
   bundle_header_para->receiver_timestamp=ntohs(bundle_header.receiver_timestamp);
   strcpy( bundle_header_para->sender_id, inet_ntoa(bundle_header.sender_id) );
   bundle_header_para->sender_timestamp=ntohs(bundle_header.sender_timestamp);
   bundle_header_para->sn=ntohs(bundle_header.bundle_sn);
   bundle_header_para->x_supp=bundle_header.x_supp_man*
                                                   pow(10, bundle_header.x_supp_exp);
   bundle_header_para->dsn_count=bundle_header.dsn_count;
   bundle_header_para->len=ntohs(bundle_header.len);

   return 0;
}

/**************************************************************************************/
/* given struct feedback_header_t, fill in struct feedback_hdr_t, which is the actual */
/* feedback header used during transmission                                           */
/**************************************************************************************/
int assemble_feedback_header(struct feedback_header_t feedback_header_para,
                                          struct feedback_hdr_t * feedback_header)
{  int a, b, m;
#ifndef UNIX
   __u8 * byte_pointer;
   __u8 temp_byte1;
#endif
   feedback_header->version= 0x2;
   feedback_header->type= 0x1;
   feedback_header->fb_nr=feedback_header_para.fb_nr;
   feedback_header->receiver_leave=feedback_header_para.receiver_leave_flag;
   feedback_header->have_loss=feedback_header_para.have_loss_flag;
   feedback_header->have_rtt=feedback_header_para.have_rtt_flag;
   feedback_header->sender_timestamp=htons(feedback_header_para.sender_timestamp);
   feedback_header->receiver_timestamp=htons(feedback_header_para.receiver_timestamp);
   feedback_header->sender_id.s_addr=inet_addr(feedback_header_para.sender_id);
   feedback_header->receiver_id.s_addr=inet_addr(feedback_header_para.receiver_id);

   /* convert x_r from a double into 16bit */
   m=(int) (log(feedback_header_para.x_r)/log(10) + 1);
   a=(int) (feedback_header_para.x_r/pow(10,(m-4)));
   b=m-4;
   if(b<0)
   {  a=(int)feedback_header_para.x_r;
      b=0;
   }
   else if(a>=2048)
   {  a=a/10;
      b=b+1;
   }
   feedback_header->x_r_man=a;
   feedback_header->x_r_exp=b;

#ifndef UNIX
   /* transfrom the bit sequence from windows to UNIX */
   byte_pointer=(__u8 *)feedback_header;
   temp_byte1=*(byte_pointer+2);
   *(byte_pointer+2)=*(byte_pointer+3);
   *(byte_pointer+3)=temp_byte1;
#endif
   
   return 0;
}

/**************************************************************************************/
/* given struct feedback_hdr_t, fill in struct feedback_header_t, which will be used  */
/* as parameter when TFMCC processing feedback information                            */
/**************************************************************************************/
int dessemble_feedback_header(struct feedback_header_t * feedback_header_para, 
                                             struct feedback_hdr_t feedback_header)
{
#ifndef UNIX
   __u8 * byte_pointer;
   __u8 temp_byte1;
   /* transform from Unix bit sequency into Windows sequency */
   byte_pointer=(__u8 *)&feedback_header;
   temp_byte1=*(byte_pointer+2);
   *(byte_pointer+2)=*(byte_pointer+3);
   *(byte_pointer+3)=temp_byte1;
#endif


   feedback_header_para->fb_nr=feedback_header.fb_nr;
   feedback_header_para->receiver_leave_flag=feedback_header.receiver_leave;
   feedback_header_para->have_loss_flag=feedback_header.have_loss;
   feedback_header_para->have_rtt_flag=feedback_header.have_rtt;
   feedback_header_para->x_r=feedback_header.x_r_man*pow(10, feedback_header.x_r_exp);
   feedback_header_para->sender_timestamp=ntohs(feedback_header.sender_timestamp);
   feedback_header_para->receiver_timestamp=ntohs(feedback_header.receiver_timestamp);
   strcpy(feedback_header_para->sender_id, inet_ntoa(feedback_header.sender_id));
   strcpy(feedback_header_para->receiver_id, inet_ntoa(feedback_header.receiver_id));

   return 0;
}

/**************************************************************************************/
/*send out a feedback through socket                                                  */
/**************************************************************************************/
int send_feedback(unsigned int sock_fd, unsigned long int dest_mca, 
                                             struct feedback_hdr_t feedback_header)
{  char buffer[FEEDBACK_HEADER_SIZE+5];
   struct sockaddr_in sa;

   int n;
   memcpy(buffer, &feedback_header, FEEDBACK_HEADER_SIZE);

   printf("send feedback\n");
   memset(&sa, 0, sizeof(sa));
   sa.sin_family = AF_INET;
   sa.sin_addr.s_addr=dest_mca;
   sa.sin_port = htons(8000);


   /* printf( "----- send feedback at time %d\n", get_timestamp()); */
   n=sendto(sock_fd, buffer, FEEDBACK_HEADER_SIZE, 0,(struct sockaddr*)&sa, sizeof(sa));

   if (n < 0)
   {  printf("feedback send error \n\a");
      exit(1);
   }
   /* printf("n=%d\n", n); */

   return 0;
}

/******************************************************************************/
/*   seach in tfmcc_mca[] to find the index (synchronized by tfmcc_mca_mutex) */
/*   this index is used to point to a record in group_info[]                  */
/*   if the mca is not found in the tfmcc_mca[]                               */
/*       create a new record for the new mca, and create corresponding thread */
/*   if no empty slot found, then return -1                                   */
/*   else, return the index                                                   */
/*   this function should be called in SRMP whenever register or subscribe a  */
/*     new mca                                                                */
/******************************************************************************/
int search_tfmcc_mca(unsigned long int mca, unsigned int socket_id)
{  int i;
   int * group_ind;
   struct sock_ind_t * para;
#ifdef UNIX
   pthread_t thread_id;
   pthread_mutex_lock(&(tfmcc_mca_mutex));
#else
   WaitForSingleObject(tfmcc_mca_mutex, INFINITE);
#endif
   for(i=0; i<20; i=i+1)
   {  if(tfmcc_mca[i]==mca)   /* found the specified mca */
         break;
      if(tfmcc_mca[i]==0)     /* no found */
      {  tfmcc_mca[i]=mca;

         /* create new thread for the new mca group */
         group_ind=(int *)malloc(sizeof(int));
         para=(struct sock_ind_t *)malloc(sizeof(struct sock_ind_t));
         *group_ind=i;
         if(socket_id<=0)  /* upon_send_bundle(), upon_receive_bundle(),
                            upon_receive_feedback(), will call with sock_id=0 */
         {  printf("TFMCC thread was not started when sending or receiving bundle\a");
            return(-1);
         }
         para->socket_id=socket_id;
         para->group_index=i;

#ifdef UNIX
         pthread_create(&thread_id, NULL, thread_fb_nr, (void *)group_ind);
         pthread_create(&thread_id, NULL, timer_thread, (void *)para);
#else
         _beginthread(thread_fb_nr,0, (void *)group_ind);
         _beginthread(timer_thread,0, (void *) para);
#endif
         break;
      }
   }

   if(i<=20)
   {  
#ifdef UNIX
      pthread_mutex_unlock(&(tfmcc_mca_mutex));
#else
      ReleaseMutex(tfmcc_mca_mutex);
#endif
      return i;
   }
   else
   {  printf("no empty slot for the new mutlcast group\a\n");
      exit(-1);
   }

}



double max_value(double a, double b)
{  if(a<b)
      return b;
   else
      return a;
}

double min_value(double a, double b)
{  if(a<b)
      return a;
   else
      return b;
}


