/*----------------------------------------------------------------*
|tfmcc_proto.h                                                    |
|@version: 2.0                                                    |
|Last Modification: Dec 14th, 2003 , by Fei, based on 0516032252  |
|                                                                 |
| Define TFMCC constants, data structures and function prototypes |
*----------------------------------------------------------------*/
/*----------------------------------------------------------------*
|    Copyright 2001-2003 Networking and Simulation Laboratory     |
|         George Mason University, Fairfax, Virginia              |
|                                                                 |
| Permission to use, copy, modify, and distribute this            |
| software and its documentation for academic purposes is hereby  |
| granted without fee, provided that the above copyright notice   |
| and this permission appear in all copies and in supporting      |
| documentation, and that the name of George Mason University     |
| not be used in advertising or publicity pertaining to           |
| distribution of the software without specific, written prior    |
| permission. GMU makes no representations about the suitability  |
| of this software for any purposes.  It is provided "AS IS"      |
| without express or implied warranties.  All risk associated     |
| with use of this software is expressly assumed by the user.     |
 *----------------------------------------------------------------*/



#define Q_HIS 0.5
#define Q_HIS_CLR 0.9
#define INITIAL_RMAX 50
// the mean time interval to calculate sending rate
#define MEAN_SEND_PERIOD 300
// the mean time interval to calculate receiving rate
#define MEAN_RECV_PERIOD 300
#define DEFAULT_MEAN_PKT_SIZE 500

#define BUNDLE_HEADER_SIZE 24
#define FEEDBACK_HEADER_SIZE 16
#define SIMU_DELAY 0   // a artifical dealy
#define MAXIMUM_LOSS_PER_EVENT 5

typedef unsigned char __u8;
typedef unsigned short __u16;
typedef unsigned int __u32;
typedef unsigned long __u64;

char ip_address[20];
double weight[8]={0.2, 0.4, 0.6, 0.8, 1, 1, 1, 1};

// this is the bundle header for transmission
#ifdef UNIX
struct bundle_hdr_t
{  __u8  version:    4,   
         type:       4;
   __u8  fb_nr:      4,
         reserved:   3,
		 is_clr_flag:1;
   __u16 bundle_sn;
   struct in_addr sender_id;
   struct in_addr receiver_id;
   __u16 sender_timestamp;
   __u16 receiver_timestamp;
   __u16 x_supp_exp: 5,
         x_supp_man: 11;
   __u16 r_max;
   __u8  dsn_count;
   __u8  pad;
   __u16 len;
};


// this is the feedback header for transmission
struct feedback_hdr_t {
	__u8 version:4,
			type:4;
	__u8 fb_nr:4,
			reserved:1,
			receiver_leave:1,
			have_loss:1,
			have_rtt:1;
	__u16 	x_r_exp:5,
	        x_r_man:11;
	__u16 sender_timestamp;
	__u16 receiver_timestamp;
	struct in_addr sender_id;
	struct in_addr receiver_id;
};
#else
struct bundle_hdr_t
{  __u8  type:       4,
         version:    4;
   __u8  is_clr_flag:1,
         reserved:   3,
            fb_nr:   4;
   __u16 bundle_sn;
   struct in_addr sender_id;
   struct in_addr receiver_id;
   __u16 sender_timestamp;
   __u16 receiver_timestamp;
   __u16 x_supp_man: 11,
         x_supp_exp: 5;
   __u16 r_max;
   __u8  dsn_count;
   __u8  pad;
   __u16 len;
};


// this is the feedback header for transmission
struct feedback_hdr_t {
	__u8 	type:4,
            version:4;
	__u8    have_rtt:1,
			have_loss:1,
			receiver_leave:1,
			reserved:1,
            fb_nr:4;
	__u16 x_r_man:11,
			x_r_exp:5;
	__u16 sender_timestamp;
	__u16 receiver_timestamp;
	struct in_addr sender_id;
	struct in_addr receiver_id;
};
#endif
// note: the tfmcc draft, if there is no feedback round in slow start, then which receiver is the slowest receiver?
//     if there is, then during the first feedback round, say 4*500ms=2second, then sending rate should keep 1 packet per second
//     eigher case is a problem!




//  this is the bundle header for parameter pasing between SRMP and TFMCC
//  the data structure for a bundle header
struct bundle_header_t {
	char sender_id[20];
	char receiver_id[20];
	unsigned short int sender_timestamp;
	unsigned short int receiver_timestamp;
	unsigned short int received_time;
	double x_supp;
	unsigned short int r_max;
	int is_clr_flag;
	int fb_nr;
	int sn;
	int dsn_count;
	int len;
};

// the data structure for a feedback header
struct feedback_header_t{
	double x_r;
	unsigned short int sender_timestamp;

	char sender_id[20];
	char receiver_id[20];
	int fb_nr;
	unsigned short int receiver_timestamp;
	unsigned short int received_time;
	int have_rtt_flag;
	int have_loss_flag;
	int receiver_leave_flag;
};


// the data structure for feedback infromation from receivers.
struct feedback_list_t{
	char receiver_id[20];
	int fb_nr;
	unsigned short int receiver_timestamp;
	unsigned short int received_time;
	unsigned short int rtt;  // note, rtt? maybe not necessary
	int have_rtt_flag;
	int have_loss_flag; // note, maybe not necessary
	int receiver_leave_flag;// note, maybe not necessary
	int echoed_flag; // note maybe not necessary
	int is_clr_flag;  // used by the receiver to indicate its current status.

	struct feedback_list_t * next;
	struct feedback_list_t * prev;
};


// the data structure for sender
// note, this should be an array instead of one element,
// becasue each SRTP instance may act as multiple senders.

struct packet_history_t{
	int sn;
	int len;
	unsigned short int received_time;
};

// data structure for the receiver to keep track of each sender.
struct other_sender_t{
	// note: this receiver_id equals the local sender_id, so not necessary!
	char receiver_id[20];  // note to initialize receiver_id


	int s_loss;  // the sn of the most recent packet that has be determined to be lost
	unsigned short int t_loss; // the nominal arrival time of the lost packet above
	int s_old;   // the sn of the packet that started the most recent lost event
	unsigned short int t_old;  // the arrival time of the packet with sn=s_old
	int have_loss_flag;
	//int rtt=100;, use the rtt inside feedback_list

	struct packet_history_t packet_history[20]; // packet history, 5 most recent packet
	int loss_interval[10]; // 10 most recent loss interval. does not include that since the most recent loss event.
							// note to reset all loss_interval to 0


	int is_clr_flag;
	//unsigned short int rtt;
	struct bundle_header_t sender_list; // for each sender, there will be an entry
	struct feedback_list_t feedback_list; // for each sender, there will be an entry
	double x_recv;
	double mean_packet_size;
	char clr_id[20];
	double minimum_x_supp; // the minimum x_r from all other non-clr receiver's feedback
	int leave_rtt_count;

	int number_of_loss;	// the number of packet loss in current loss event, by now.

};

// feedback timer link list
struct time_event_t{
	unsigned short int active_time;  // the time stamp when the current time event expire
	int fb_nr; // the fb_nr when set up the timer
	int index;  // index to receiver_info[20]
	struct time_event_t * next;
};


// the data struct for storing sending side message history
//     in order to calculate mean sending rate during MEAN_SEND_PERIOD
struct srmp_send_his_t{
	unsigned short int sending_time;
	int len;
	int mode;
};

// the data structure for a sender
// include: basic sender information for the group
//          CLR information
//          a linked list for all the feedback received from receivers
struct group_info_t{
	unsigned long int mca;

	char sender_id[20];
	int fb_nr;
	int sn;
	double x_supp;
	double x_send;
	double mean_packet_size;
	unsigned short int r_max;
	struct feedback_list_t * feedback_listhead;
	struct feedback_list_t * feedback_listtail;
	int clr_new_flag;
	int clr_leave_flag;
	int in_slowstart_flag;
	int no_large_rmax; // note 0,2, this should be large_rmax, instead of no_large_rmax
	int feedback_received_flag;  //indicate a non-CLR feedback during the feedback round
	int non_feedback;  // count to 10; indicate any kind of feedback
	int clr_non_feedback; // count to 10;
//  char clr[20];
	struct feedback_list_t  clr;

	// in slow start stage: increase x_r to target_rate during one R_max
	// when out of slow start stage, "increase to x_r at a rate no more than 8s/R_max per R_max"
	double target_rate;
	unsigned short int start_point;
	double increase_rate; // for slow start only, because when out of slow start, increase rate
						// increase is 8S/R_max

	// to permit multiple receiver send back feedback rate,
	// and choose the smalleset one as the CLR
	double feedback_rate;

	// in slow start stage, slowstart_rate record the slowest rate reported
	// during one feedback round
	double slowstart_rate;
	struct feedback_list_t target_clr;

	int cancel_flag; // the flag indicates if the feedback event in the list head should be cancelled
	int current_receiver_leave; // the flag indicates if the receiver is leaving.
#ifdef UNIX
   long last_packet_time;
#else
   unsigned long int last_packet_time; // in case sending rate is: 1 packet per 64 seconds
#endif


	struct other_sender_t other_senders[20];
	struct srmp_send_his_t srmp_send_his[20];



	struct time_event_t * time_event_head;
	struct time_event_t * time_event_tail;
#ifdef UNIX
	pthread_cond_t time_event;
	pthread_mutex_t other_senders_mutex;
	pthread_cond_t fbnr_event; // SRMP will wait on this event, when slow start, packet rate=1packet/r_max
	pthread_mutex_t time_event_mutex;
	pthread_mutex_t mutex;
	pthread_cond_t event;
#else
	HANDLE time_event;
	HANDLE other_senders_mutex;
	HANDLE fbnr_event; // SRMP will wait on this event, when slow start, packet rate=1packet/r_max
	HANDLE time_event_mutex;
	HANDLE mutex;
	HANDLE event;
#endif

};

struct sock_ind_t{
	unsigned int socket_id;
	int group_index;
};


#ifdef UNIX
// the thread for sender to control feedback round
void * thread_fb_nr(void * index);
// the thread for all the receivers, to setup timer and triger feedback after timer expired
void * timer_thread(void * mp);

#else
// the thread for sender to control feedback round
void thread_fb_nr(void * index);
// the thread for all the receivers, to setup timer and triger feedback after timer expired
void timer_thread(void * mp);
#endif



// the function to initialze data structe and create threads
void initialize_tfmcc();

// the function to generate time stamp
unsigned short int get_timestamp();


int upon_send_bundle( struct bundle_header_t * bundler_header, double * sending_rate, 
					 double packet_size, unsigned long int mca);
int upon_receive_feedback(struct feedback_header_t feedback_header1, unsigned long int mca);
int upon_receive_bundle(struct bundle_header_t bundle_header, double packet_size,
						unsigned long int mca);


int receiver_leave_group(unsigned long int mca);
int receiver_resume_group(unsigned long int mca);

int search_tfmcc_mca(unsigned long int mca, unsigned int socket_id);


// function to update sending packet history.
// this function should be called each time before SRMP tries to put a message
// into the bundle. 
// return value: 0 means packet should NOT be dropped
//               1 means packet should be dropped

int drop_mode_0(int mode, int len, unsigned long int mca);


// functions to assemble and dessemble feedback and bundle header
int dessemble_feedback_header(struct feedback_header_t * feedback_header_para, struct feedback_hdr_t feedback_header);
int assemble_feedback_header(struct feedback_header_t feedback_header_para, struct feedback_hdr_t * feedback_header);
int dessemble_bundle_header(struct bundle_header_t * bundle_header_para, struct bundle_hdr_t bundle_header);
int assemble_bundle_header(struct bundle_header_t bundle_header_para, struct bundle_hdr_t * bundle_header);
void show_bundle_header(struct bundle_header_t * bundle_header);
void show_feedback_header(struct feedback_header_t  feedback_header);

// function to sendout a feedback through the socket
int send_feedback(unsigned int sock_fd, unsigned long int mca, struct feedback_hdr_t feedback_header);

double max_value(double a, double b);
double min_value(double a, double b);